from robotframework_ls.impl.protocols import (
    IKeywordArg,
    IRobotToken,
)
from typing import Optional, List, Deque, Iterator, Dict, Union, Sequence
import itertools
from robocorp_ls_core.lsp import Error
from robocorp_ls_core.constants import Null, NULL


class _Match:
    def __init__(
        self, definition_arg: IKeywordArg, token_definition_id_to_index: Dict[int, int]
    ):
        self._definition_arg = definition_arg
        self._token_definition_id_to_index = token_definition_id_to_index

    def get_active_parameter_in_definition(self):
        return self._token_definition_id_to_index.get(id(self._definition_arg), -1)


class SkipAnalysisControlFlowException(Exception):
    pass


class UsageInfoForKeywordArgumentAnalysis:
    def __init__(self, node, token_to_report_missing_argument):
        self.node = node
        self._token_to_report_missing_argument = token_to_report_missing_argument

    def get_token_to_report_argument_missing(self):
        return self._token_to_report_missing_argument


class KeywordArgumentAnalysis:
    def __init__(self, keyword_args: Sequence[IKeywordArg]) -> None:

        args = self._keyword_args = keyword_args

        self.found_star_arg: Optional[IKeywordArg] = None
        self.found_keyword_arg: Optional[IKeywordArg] = None
        self._star_arg_index = -1
        self._keyword_arg_index = -1

        for i, arg in enumerate(args):
            if arg.is_star_arg:
                self.found_star_arg = arg
                self._star_arg_index = i

            elif arg.is_keyword_arg:
                self.found_keyword_arg = arg
                self._keyword_arg_index = i

    def _compute_active_parameter_fallback(
        self,
        usage_info_argument_index: int,
        usage_info_arg_tokens: Sequence[IRobotToken],
    ) -> int:
        """
        If we didn't have an exact match (because the current call would
        probably be inconsistent as the user may be typing it), provide a
        fallback which works better is such situations.
        """
        from robotframework_ls.impl.text_utilities import is_variable_text
        from robotframework_ls.impl.variable_resolve import find_split_index

        if usage_info_argument_index < 0:
            return -1

        if usage_info_argument_index >= len(usage_info_arg_tokens):
            # This happens when the user is starting to type an argument.
            return usage_info_argument_index

        active_parameter: int = usage_info_argument_index
        if self._keyword_arg_index == -1 and self._star_arg_index >= 0:
            if active_parameter >= self._star_arg_index:
                return self._star_arg_index

        caller_arg_value = usage_info_arg_tokens[active_parameter].value
        definition_keyword_args = self._keyword_args
        # Now, it's also possible that we're dealing with an assign here... let's
        # see if this is the case.
        eq: int = find_split_index(caller_arg_value)
        if eq != -1:
            name = caller_arg_value[:eq]
            for i, keyword_arg in enumerate(definition_keyword_args):
                arg_name = keyword_arg.original_arg
                if is_variable_text(arg_name):
                    arg_name = arg_name[2:-1]
                arg_name = arg_name
                if name == arg_name:
                    active_parameter = i
                    break
            else:
                # We do NOT have a match (match keyword arg / star arg if present...)
                if self._keyword_arg_index >= 0:
                    active_parameter = self._keyword_arg_index

                elif self._star_arg_index >= 0:
                    active_parameter = self._star_arg_index

                else:
                    # This is actually off (error in call).
                    active_parameter = -1

        else:
            saw_eq: bool = False
            for arg in usage_info_arg_tokens[:active_parameter]:
                saw_eq = "=" in arg.value
                if saw_eq:
                    break

            if saw_eq and self._keyword_arg_index >= 0:
                return self._keyword_arg_index

            # Ok, does not have an assign, let's inspect the original signature
            # to detect where this should be put there (positional arg or
            # stararg).
            for i, definition_arg in enumerate(definition_keyword_args):
                if i == active_parameter:
                    break

                if definition_arg.is_star_arg:
                    active_parameter = i
                    break

                if definition_arg.is_keyword_arg:
                    # This is actually off (error in call).
                    active_parameter = -1
                    break

        return active_parameter

    def _iter_args(self, tokens):
        from robot.api import Token

        for token in tokens:
            if token.type == Token.ARGUMENT:
                if token.value.startswith("&{") or token.value.startswith("@{"):
                    # All bets are off in this case (it may match anything...)
                    raise SkipAnalysisControlFlowException()

                yield token

    def _collect_keyword_usage_errors_and_build_definition_map(
        self,
        usage_info: UsageInfoForKeywordArgumentAnalysis,
        usage_token_id_to_definition_arg_match: Union[Dict[int, _Match], Null] = NULL,
        collect_errors=True,
    ) -> Iterator[Error]:
        try:
            yield from self._collect_keyword_usage_errors_and_build_definition_map_raises_exc(
                usage_info, usage_token_id_to_definition_arg_match, collect_errors
            )
        except SkipAnalysisControlFlowException:
            pass

    def _collect_keyword_usage_errors_and_build_definition_map_raises_exc(
        self,
        usage_info: UsageInfoForKeywordArgumentAnalysis,
        usage_token_id_to_definition_arg_match: Union[Dict[int, _Match], Null] = NULL,
        collect_errors=True,
    ) -> Iterator[Error]:
        """
        In this function we build the contents of usage_token_id_to_definition_arg_match
        and collect the errors (if collect_errors=True).
        """
        from robotframework_ls.impl.ast_utils import create_error_from_node
        from collections import deque
        from robotframework_ls.impl.text_utilities import is_variable_text
        from robotframework_ls.impl.variable_resolve import find_split_index

        # Pre-requisite.
        keyword_token = usage_info.get_token_to_report_argument_missing()
        if not keyword_token:
            return

        # deque (initially with all args -- args we match are removed
        # as we go forward).
        definition_keyword_args_deque: Deque[IKeywordArg] = deque()

        # id(arg) -> index in the definition.
        token_definition_id_to_index: Dict[int, int] = {}

        # Contains all names we can match -> the related keyword arg.
        definition_keyword_name_to_arg: Dict[str, IKeywordArg] = {}

        # The ones that are matched are filled as we go.
        definition_arg_matched: Dict[IKeywordArg, bool] = {}

        # Fill our basic structures.
        for i, definition_arg in enumerate(self._keyword_args):
            definition_keyword_args_deque.append(definition_arg)
            token_definition_id_to_index[id(definition_arg)] = i

            if definition_arg.is_star_arg:
                # Skip not matched by name
                continue
            if definition_arg.is_keyword_arg:
                # Skip not matched by name
                continue

            arg_name = definition_arg.arg_name
            if is_variable_text(arg_name):
                arg_name = arg_name[2:-1]

            definition_keyword_name_to_arg[arg_name] = definition_arg

        tokens_args_to_iterate = self._iter_args(usage_info.node.tokens)
        # Fill positional args
        for token_arg in tokens_args_to_iterate:
            if not definition_keyword_args_deque:
                # No more arguments to consume...
                # Add it back as it still wasn't consumed (this is an error).
                tokens_args_to_iterate = itertools.chain(
                    iter([token_arg]), tokens_args_to_iterate
                )
                break

            eq_index = find_split_index(token_arg.value)

            if eq_index == -1:
                matched_keyword_arg: IKeywordArg = (
                    definition_keyword_args_deque.popleft()
                )
                if matched_keyword_arg.is_star_arg:
                    # Add star arg back as we may keep on matching it.
                    definition_keyword_args_deque.appendleft(matched_keyword_arg)
                    usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                        matched_keyword_arg, token_definition_id_to_index
                    )
                    continue

                if matched_keyword_arg.is_keyword_arg:
                    if collect_errors:
                        error = create_error_from_node(
                            usage_info.node,
                            f"Unexpected positional argument: {token_arg.value}",
                            tokens=[token_arg],
                        )
                        yield error

                    # Add it (just because the user may be typing it...)
                    usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                        matched_keyword_arg, token_definition_id_to_index
                    )

                    # Finish as it's inconsistent now.
                    return

                definition_arg_matched[matched_keyword_arg] = True
                usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                    matched_keyword_arg, token_definition_id_to_index
                )

            else:
                if self.found_keyword_arg is not None:
                    # Something with '=' always matches keyword args, even if
                    # no named args would be matched. Add it back to go to
                    # the part where we match arguments by name.
                    tokens_args_to_iterate = itertools.chain(
                        iter([token_arg]), tokens_args_to_iterate
                    )
                    break

                name = token_arg.value[:eq_index]
                found_definition_arg = definition_keyword_name_to_arg.get(name, None)
                if found_definition_arg is not None:
                    if definition_arg_matched.get(found_definition_arg):
                        error = create_error_from_node(
                            usage_info.node,
                            f"Multiple values for argument: {name}",
                            tokens=[token_arg],
                        )
                        yield error
                        return

                    # First with eq (named argument) that matched something. Add
                    # it back to go to the part where we match arguments by name.
                    tokens_args_to_iterate = itertools.chain(
                        iter([token_arg]), tokens_args_to_iterate
                    )
                    break

                matched_keyword_arg = definition_keyword_args_deque[0]
                usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                    matched_keyword_arg, token_definition_id_to_index
                )

                if matched_keyword_arg.is_star_arg:
                    # Special-case, if the last thing we have is a star-arg, it'll
                    # consume everything (even equals) as long as no named arguments
                    # were matched first.
                    pass
                else:
                    # Matched some argument (the '=' became a part of the value).
                    definition_arg_matched[matched_keyword_arg] = True
                    definition_keyword_args_deque.popleft()

        if not definition_keyword_args_deque:
            if collect_errors:
                # If we have no more args to consume, everything else is an error.
                for token_arg in tokens_args_to_iterate:
                    if collect_errors:
                        error = create_error_from_node(
                            usage_info.node,
                            f"Unexpected argument: {token_arg.value}",
                            tokens=[token_arg],
                        )
                        yield error
            return

        # Ok, from this point onwards we need to match only by name / stararg / keyword_token arg

        # Now, consume all the ones given by name.
        for token_arg in tokens_args_to_iterate:
            eq_index = find_split_index(token_arg.value)
            if eq_index >= 0:
                name = token_arg.value[:eq_index]
                found_definition_arg = definition_keyword_name_to_arg.pop(name, None)
                if not found_definition_arg:
                    if self.found_keyword_arg is not None:
                        usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                            self.found_keyword_arg, token_definition_id_to_index
                        )
                    else:
                        if collect_errors:
                            error = create_error_from_node(
                                usage_info.node,
                                f"Unexpected named argument: {name}",
                                tokens=[token_arg],
                            )
                            yield error

                else:
                    usage_token_id_to_definition_arg_match[id(token_arg)] = _Match(
                        found_definition_arg, token_definition_id_to_index
                    )

            else:
                if collect_errors:
                    error = create_error_from_node(
                        usage_info.node,
                        f"Positional argument not allowed after named arguments: {token_arg.value}",
                        tokens=[token_arg],
                    )
                    yield error
                return

        if collect_errors:
            # To finish, give errors on unmatched arguments.
            for (
                definition_name,
                definition_arg,
            ) in definition_keyword_name_to_arg.items():
                if (
                    not definition_arg.is_default_value_set()
                    and not definition_arg_matched.get(definition_arg)
                ):
                    error = create_error_from_node(
                        usage_info.node,
                        f"Mandatory argument missing: {definition_name}",
                        tokens=[keyword_token],
                    )
                    yield error

    # --------------------------------------------------------------- Public API

    def compute_active_parameter(
        self, usage_info: UsageInfoForKeywordArgumentAnalysis, lineno: int, col: int
    ) -> int:

        token_to_report_argument_missing = (
            usage_info.get_token_to_report_argument_missing()
        )
        if token_to_report_argument_missing.lineno - 1 > lineno or (
            token_to_report_argument_missing.lineno - 1 == lineno
            and token_to_report_argument_missing.end_col_offset >= col
        ):
            return -1

        from robot.api import Token

        usage_info_argument_index: int = 0
        # We need to find out the current arg/separator.
        after_last_arg: List[Token] = []
        usage_info_arg_tokens: List[Token] = []

        for token in usage_info.node.tokens:
            if token.type == Token.ARGUMENT:
                usage_info_arg_tokens.append(token)
                usage_info_argument_index += 1
                del after_last_arg[:]

            elif token.type in (Token.SEPARATOR, Token.EOL, Token.EOS):
                after_last_arg.append(token)
            else:
                # Keyword name token
                del after_last_arg[:]

            if token.lineno - 1 == lineno:
                if (token.end_col_offset - 1) >= col:
                    break

        if token.type == Token.ARGUMENT:
            usage_info_argument_index -= 1

        elif after_last_arg:
            # Check if we are in prev/next based on the number of spaces found
            # up to the current cursor position.
            # i.e.: in `Call  arg ` we still need to have an usage_info_argument_index == 0
            # i.e.: in `Call  arg  ` we need to have an usage_info_argument_index == 1
            whitespaces_found = []
            if token.lineno - 1 == lineno:
                if (token.end_col_offset - 1) <= col:
                    whitespaces_found.append(token.value)
                else:
                    whitespaces_found.append(
                        token.value[: -(token.end_col_offset - col)]
                    )
            s = "".join(whitespaces_found)
            if len(s) <= 1:
                usage_info_argument_index -= 1

        if usage_info_argument_index == -1:
            return -1

        active_parameter: int = -1

        try:
            arg_token = usage_info_arg_tokens[usage_info_argument_index]
        except IndexError:
            pass
        else:
            if arg_token is not None:
                token_id_to_match: Dict[int, _Match] = {}
                for (
                    _error
                ) in self._collect_keyword_usage_errors_and_build_definition_map(
                    usage_info, token_id_to_match, collect_errors=False
                ):
                    pass

                matched = token_id_to_match.get(id(arg_token))
                if matched is not None:
                    active_parameter = matched.get_active_parameter_in_definition()

        # Something didn't work out...
        if active_parameter == -1:
            return self._compute_active_parameter_fallback(
                usage_info_argument_index, usage_info_arg_tokens
            )
        return active_parameter

    def collect_keyword_usage_errors(
        self,
        usage_info: UsageInfoForKeywordArgumentAnalysis,
    ) -> Iterator[Error]:
        yield from self._collect_keyword_usage_errors_and_build_definition_map(
            usage_info
        )



from robot.api.parsing import Token, ModelTransformer, SectionHeader, EmptyLine
from robot.parsing.model.statements import Statement

from robotidy.exceptions import InvalidParameterValueError

try:
    from robot.api.parsing import InlineIfHeader
except ImportError:
    InlineIfHeader = None


class MergeAndOrderSections(ModelTransformer):
    """
    Merge duplicated sections and order them.

    Default order is: Comments > Settings > Variables > Test Cases > Keywords.

    You can change sorting order by configuring ``order`` parameter with comma separated list of section names (without
    spaces)::

        robotidy --transform MergeAndOrderSections:order=settings,keywords,variables,testcases,comments

    Because merging and changing the order of sections can shuffle your empty lines it's greatly advised to always
    run ``NormalizeNewLines`` transformer after this one.

    If both ``*** Test Cases ***`` and ``*** Tasks ***`` are defined in one file they will be merged into one (header
    name will be taken from first encountered section).

    Any data before first section is treated as comment in Robot Framework. This transformer add ``*** Comments ***``
    section for such lines::

        i am comment
        # robocop: disable
        *** Settings ***

    To::

        *** Comments ***
        i am comment
        # robocop: disable
        *** Settings ***

    You can disable this behaviour by setting ``create_comment_section`` to False.

    See https://robotidy.readthedocs.io/en/latest/transformers/MergeAndOrderSections.html for more examples.
    """

    def __init__(self, order: str = "", create_comment_section: bool = True):
        self.sections_order = self.parse_order(order)
        self.create_comment_section = create_comment_section

    def parse_order(self, order):
        default_order = (
            Token.COMMENT_HEADER,
            Token.SETTING_HEADER,
            Token.VARIABLE_HEADER,
            Token.TESTCASE_HEADER,
            Token.KEYWORD_HEADER,
        )
        if not order:
            return default_order
        parts = order.lower().split(",")
        map = {
            "comments": Token.COMMENT_HEADER,
            "comment": Token.COMMENT_HEADER,
            "settings": Token.SETTING_HEADER,
            "setting": Token.SETTING_HEADER,
            "variables": Token.VARIABLE_HEADER,
            "variable": Token.VARIABLE_HEADER,
            "testcases": Token.TESTCASE_HEADER,
            "testcase": Token.TESTCASE_HEADER,
            "keywords": Token.KEYWORD_HEADER,
            "keyword": Token.KEYWORD_HEADER,
        }
        parsed_order = []
        for part in parts:
            parsed_order.append(map.get(part, None))
        if any(header not in parsed_order for header in default_order) and len(parsed_order) != len(default_order):
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "order",
                order,
                "Custom order should be provided in comma separated list with all section names:\n"
                "order=comments,settings,variables,testcases,variables",
            )
        return parsed_order

    def visit_File(self, node):  # noqa
        if len(node.sections) < 2:
            return node
        sections = {}
        last = len(node.sections) - 1
        for index, section in enumerate(node.sections):
            if index == last:
                section = self.from_last_section(section)
            section_type = self.get_section_type(section)
            if section_type not in sections:
                sections[section_type] = section
            else:
                if len(section.header.data_tokens) > 1:
                    print(
                        f"{node.source}: Merged duplicated section has section header comments. "
                        "Only header comments from first section header of the same type are preserved."
                    )
                sections[section_type].body += section.body
        node.sections = [sections[order] for order in self.sections_order if order in sections]
        return node

    @staticmethod
    def normalize_eol(tokens):
        new_tokens = []
        for tok in tokens:
            if tok.type == Token.EOL:
                tok.value = "\n"
            new_tokens.append(tok)
        return new_tokens

    def from_last_section(self, node):
        """Last node use different logic for new line marker. It is not possible to preserve all empty lines but
        we need at least ensure that following code::

             *** Test Case ***
             *** Variables ***

        Will not become::
            *** Variables ****** Test Case ***

        """
        # Empty sections, just *** Test Cases *** etc
        if not node.body:
            node.header.tokens = self.normalize_eol(node.header.tokens)
            return node

        # Settings, Variables or Comments
        if not hasattr(node.body[-1], "body"):
            node.body[-1].tokens = self.normalize_eol(node.body[-1].tokens)
            return node

        # Last keyword or test case
        if not node.body[-1].body:
            node.body[-1].body.append(EmptyLine.from_params())
            return node

        last_statement = node.body[-1].body[-1]
        if hasattr(last_statement, "end"):
            if (
                InlineIfHeader
                and hasattr(last_statement, "header")
                and isinstance(last_statement.header, InlineIfHeader)
            ):
                if not last_statement.errors:
                    node.body[-1].body[-1].body[-1].tokens = self.normalize_eol(last_statement.body[-1].tokens)
                return node
            if last_statement.end:  # not end if parsing errors
                node.body[-1].body[-1].end.tokens = self.normalize_eol(last_statement.end.tokens)
        else:
            node.body[-1].body[-1] = Statement.from_tokens(self.normalize_eol(last_statement.tokens))
        return node

    def get_section_type(self, section):
        header_tokens = (
            Token.COMMENT_HEADER,
            Token.TESTCASE_HEADER,
            Token.SETTING_HEADER,
            Token.KEYWORD_HEADER,
            Token.VARIABLE_HEADER,
        )
        if section.header:
            name_token = section.header.get_token(*header_tokens)
            section_type = name_token.type
        else:
            section_type = Token.COMMENT_HEADER
            if self.create_comment_section:
                section.header = SectionHeader.from_params(section_type, "*** Comments ***")
        return section_type



import sys
import threading
from typing import Optional, Any


if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass

    class TypedDict(object):
        def __init_subclass__(self, *args, **kwargs):
            pass


else:
    from typing import Protocol
    from typing import TypedDict


class ActionResultDict(TypedDict):
    success: bool
    message: Optional[
        str
    ]  # if success == False, this can be some message to show to the user
    result: Any


class EvaluateTextTypedDict(TypedDict):
    prefix: str  # The prefix added to the evaluation
    full_code: str  # The final code to be evaluated
    indent: str


class IOnReadyCall(object):
    def __call__(self, interpreter: "IRobotFrameworkInterpreter"):
        pass


class IRobotFrameworkInterpreter(Protocol):
    def compute_evaluate_text(
        self, code: str, target_type: str = "evaluate"
    ) -> EvaluateTextTypedDict:
        """
        :param target_type:
            'evaluate': means that the target is an evaluation with the given code.
                This implies that the current code must be changed to make sense
                in the given context.

            'completions': means that the target is a code-completion
                This implies that the current code must be changed to include
                all previous evaluation so that the code-completion contains
                the full information up to the current point.
        """

    def evaluate(self, code: str) -> ActionResultDict:
        pass

    def initialize(self, on_main_loop: IOnReadyCall):
        pass


class IMessage(Protocol):
    def __call__(self, interpreter: IRobotFrameworkInterpreter):
        pass

    event: threading.Event
    action_result_dict: ActionResultDict



import os
import sys
from robotframework_ls.constants import NULL
from robocorp_ls_core.robotframework_log import get_logger
import threading
from typing import Optional, Dict, Set, Iterator, Union, Any
from robocorp_ls_core.protocols import Sentinel, IEndPoint
from robotframework_ls.impl.protocols import (
    ILibraryDoc,
    ILibraryDocOrError,
    ICompletionContext,
)
import itertools
from robocorp_ls_core.watchdog_wrapper import IFSObserver
from robotframework_ls.impl.robot_lsp_constants import (
    OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS,
)
from robotframework_ls.impl.libspec_warmup import (
    _norm_filename,
    _normfile,
    LibspecWarmup,
)
from pathlib import Path
from contextlib import contextmanager
import typing
from robotframework_ls.impl.text_utilities import get_digest_from_string
from robocorp_ls_core.basic import normalize_filename

log = get_logger(__name__)


def _get_libspec_mutex_name(libspec_filename):
    from robocorp_ls_core.system_mutex import generate_mutex_name

    libspec_filename = _norm_filename(libspec_filename)
    basename = os.path.basename(libspec_filename)
    name = os.path.splitext(basename)[0]
    return generate_mutex_name(libspec_filename, prefix="%s_" % (name,))


def _get_additional_info_filename(spec_filename):
    additional_info_filename = os.path.join(spec_filename + ".m")
    return additional_info_filename


@contextmanager
def timed_acquire_mutex_for_spec_filename(spec_filename):
    from robocorp_ls_core.system_mutex import timed_acquire_mutex

    try:
        ctx = timed_acquire_mutex(_get_libspec_mutex_name(spec_filename), timeout=30)
        ctx.__enter__()
    except:
        raise RuntimeError(
            f"Unable to get mutex for: {spec_filename} after 30 seconds."
        )
    try:
        yield ctx
    finally:
        ctx.__exit__()


def _load_library_doc_and_mtime(libspec_manager, spec_filename: str, obtain_mutex=True):
    """
    :param obtain_mutex:
        Should be False if this is part of a bigger operation that already
        has the spec_filename mutex.
    """
    from robotframework_ls.impl import robot_specbuilder
    from robotframework_ls.impl.libspec_markdown_conversion import (
        load_markdown_json_version,
    )

    ctx: Any
    if obtain_mutex:
        ctx = timed_acquire_mutex_for_spec_filename(spec_filename)
    else:
        ctx = NULL
    with ctx:
        # We must load it with a mutex to avoid conflicts between generating/reading.
        try:
            mtime = os.path.getmtime(spec_filename)
            libdoc = load_markdown_json_version(libspec_manager, spec_filename, mtime)

            if libdoc is None:
                builder = robot_specbuilder.SpecDocBuilder()
                libdoc = builder.build(spec_filename)
                if libdoc.doc_format != "markdown":
                    libspec_manager.schedule_conversion_to_markdown(spec_filename)
            return libdoc, mtime
        except Exception:
            log.exception("Error when loading spec info from: %s", spec_filename)
            return None


def _load_lib_info(libspec_manager, canonical_spec_filename: str, can_regenerate: bool):
    libdoc_and_mtime = _load_library_doc_and_mtime(
        libspec_manager, canonical_spec_filename
    )
    if libdoc_and_mtime is None:
        return None
    libdoc, mtime = libdoc_and_mtime
    return _LibInfo(libdoc, mtime, canonical_spec_filename, can_regenerate)


_IS_BUILTIN = "is_builtin"
_SOURCE_TO_MTIME = "source_to_mtime"
_UNABLE_TO_LOAD = "unable_to_load"


def _create_updated_source_to_mtime(library_doc):
    sources = set()

    source = library_doc.source
    if source is not None:
        sources.add(source)

    for keyword in library_doc.keywords:
        source = keyword.source
        if source is not None:
            sources.add(source)

    source_to_mtime = {}
    for source in sources:
        try:
            # i.e.: get it before normalizing (but leave the cache key normalized).
            # This is because even on windows the file-system may end up being
            # case-dependent on some cases.
            mtime = os.path.getmtime(source)
            source = _normfile(source)
            source_to_mtime[source] = mtime
        except Exception:
            log.exception("Unable to load source for file: %s", source)
    return source_to_mtime


def _create_additional_info(
    libspec_manager, spec_filename, is_builtin, obtain_mutex=True
):
    try:
        additional_info = {_IS_BUILTIN: is_builtin}
        if is_builtin:
            # For builtins we don't have to check the mtime
            # (on a new version we update the folder).
            return additional_info

        library_doc_and_mtime = _load_library_doc_and_mtime(
            libspec_manager, spec_filename, obtain_mutex=obtain_mutex
        )
        if library_doc_and_mtime is None:
            additional_info[_UNABLE_TO_LOAD] = True
            return additional_info

        library_doc = library_doc_and_mtime[0]

        additional_info[_SOURCE_TO_MTIME] = _create_updated_source_to_mtime(library_doc)
        return additional_info

    except:
        log.exception(
            "Error creating additional info for spec filename: %s", spec_filename
        )
        return {}


def _load_spec_filename_additional_info(spec_filename):
    """
    Loads additional information given a spec filename.
    """
    import json

    try:
        additional_info_filename = _get_additional_info_filename(spec_filename)

        with open(additional_info_filename, "r") as stream:
            source_to_mtime = json.load(stream)
        return source_to_mtime
    except:
        log.exception("Unable to load source mtimes from: %s", additional_info_filename)
        return {}


def _dump_spec_filename_additional_info(
    libspec_manager, spec_filename, is_builtin, obtain_mutex=True
):
    """
    Creates a filename with additional information not directly available in the
    spec.
    """
    try:
        libspec_manager.schedule_conversion_to_markdown(spec_filename)
    except:
        log.exception("Error converting %s to markdown.", spec_filename)
    import json

    source_to_mtime = _create_additional_info(
        libspec_manager, spec_filename, is_builtin, obtain_mutex=obtain_mutex
    )
    additional_info_filename = _get_additional_info_filename(spec_filename)
    with open(additional_info_filename, "w") as stream:
        json.dump(source_to_mtime, stream, indent=2, sort_keys=True)


class _LibInfo(object):
    __slots__ = [
        "library_doc",
        "mtime",
        "_canonical_spec_filename",
        "_additional_info",
        "_invalid",
        "_can_regenerate",
    ]

    def __init__(self, library_doc: ILibraryDoc, mtime, spec_filename, can_regenerate):
        """
        :param library_doc:
        :param mtime:
        :param spec_filename:
        :param bool can_regenerate:
            False means that the information from this file can't really be
            regenerated (i.e.: this is a spec file from a library or created
            by the user).
        """
        assert library_doc
        assert mtime
        assert spec_filename

        self.library_doc = library_doc
        self.mtime = mtime

        self._can_regenerate = can_regenerate
        self._canonical_spec_filename = spec_filename
        self._additional_info = None
        self._invalid = False

    def __str__(self):
        return f"_LibInfo({self.library_doc}, {self.mtime})"

    def verify_sources_sync(self):
        """
        :return bool:
            True if everything is ok and this library info can be used. Otherwise,
            the spec file and the _LibInfo must be recreated.
        """
        if not self._can_regenerate:
            # This means that this info was generated by a library or the user
            # himself, thus, we can't regenerate it.
            return True

        if self._invalid:  # Once invalid, always invalid.
            return False

        additional_info = self._additional_info
        if additional_info is None:
            additional_info = _load_spec_filename_additional_info(
                self._canonical_spec_filename
            )
            if additional_info.get(_IS_BUILTIN, False):
                return True

            source_to_mtime = additional_info.get(_SOURCE_TO_MTIME)
            if source_to_mtime is None:
                # Nothing to validate...
                return True

            updated_source_to_mtime = _create_updated_source_to_mtime(self.library_doc)
            if source_to_mtime != updated_source_to_mtime:
                log.info(
                    "Library %s is invalid. Current source to mtime:\n%s\nChanged from:\n%s"
                    % (self.library_doc.name, source_to_mtime, updated_source_to_mtime)
                )
                self._invalid = True
                return False

        return True


class _FolderInfo(object):
    def __init__(self, folder_path, recursive):
        self.folder_path = folder_path
        self.recursive = recursive
        self.libspec_canonical_filename_to_info = {}
        self._watch = NULL
        self._lock = threading.Lock()

    def start_watch(self, observer, notifier):
        with self._lock:
            if self._watch is NULL:
                if not os.path.isdir(self.folder_path):
                    if not os.path.exists(self.folder_path):
                        log.info(
                            "Trying to track changes in path which does not exist: %s",
                            self.folder_path,
                        )
                    else:
                        log.info(
                            "Trying to track changes in path which is not a folder: %s",
                            self.folder_path,
                        )
                    return

                log.debug("Tracking folder for changes: %s", self.folder_path)
                from robocorp_ls_core.watchdog_wrapper import PathInfo

                folder_path = self.folder_path
                self._watch = observer.notify_on_any_change(
                    [PathInfo(folder_path, recursive=self.recursive)],
                    notifier.on_change,
                    (self._on_change_spec,),
                    extensions=(".py", ".libspec"),
                )

    def _on_change_spec(self, spec_file):
        with self._lock:
            spec_file_key = _norm_filename(spec_file)
            # Just add/remove that specific spec file from the tracked list.
            libspec_canonical_filename_to_info = (
                self.libspec_canonical_filename_to_info.copy()
            )
            if os.path.exists(spec_file):
                libspec_canonical_filename_to_info[spec_file_key] = None
            else:
                libspec_canonical_filename_to_info.pop(spec_file_key, None)

            self.libspec_canonical_filename_to_info = libspec_canonical_filename_to_info

    def synchronize(self):
        with self._lock:
            try:
                self.libspec_canonical_filename_to_info = self._collect_libspec_info(
                    [self.folder_path],
                    self.libspec_canonical_filename_to_info,
                    recursive=self.recursive,
                )
            except Exception:
                log.exception("Error when synchronizing: %s", self.folder_path)

    def dispose(self):
        with self._lock:
            watch = self._watch
            self._watch = NULL
            watch.stop_tracking()
            self.libspec_canonical_filename_to_info = {}

    def _collect_libspec_info(self, folders, old_libspec_filename_to_info, recursive):
        seen_libspec_files = set()
        if recursive:
            for folder in folders:
                if os.path.isdir(folder):
                    for root, _dirs, files in os.walk(folder):
                        for filename in files:
                            if filename.lower().endswith(".libspec"):
                                seen_libspec_files.add(os.path.join(root, filename))
        else:
            for folder in folders:
                if os.path.isdir(folder):
                    for filename in os.listdir(folder):
                        if filename.lower().endswith(".libspec"):
                            seen_libspec_files.add(os.path.join(folder, filename))

        new_libspec_filename_to_info: Dict[str, _LibInfo] = {}

        for filename in seen_libspec_files:
            filename = _norm_filename(filename)
            info = old_libspec_filename_to_info.get(filename)
            if info is not None:
                try:
                    curr_mtime = os.path.getmtime(filename)
                except:
                    # it was deleted in the meanwhile...
                    continue
                else:
                    if info.mtime != curr_mtime:
                        # The spec filename mtime changed, so, set to None
                        # to reload it.
                        info = None

            new_libspec_filename_to_info[filename] = info
        return new_libspec_filename_to_info


class LibspecManager(object):
    """
    Used to manage the libspec files.

    .libspec files are searched in the following directories:

    - PYTHONPATH folders                                  (not recursive)
    - Workspace folders                                   (recursive -- notifications from the LSP)
    - ${user}.robotframework-ls/specs/${python_hash}      (not recursive)

    It searches for .libspec files in the folders tracked and provides the
    keywords that are available from those (properly caching data as needed).
    """

    @classmethod
    def get_robot_version(cls) -> str:
        from robotframework_ls.impl import robot_version

        return robot_version.get_robot_version()

    @classmethod
    def get_robot_major_version(cls) -> int:
        from robotframework_ls.impl import robot_version

        return robot_version.get_robot_major_version()

    @classmethod
    def get_internal_libspec_dir(cls) -> str:
        from robotframework_ls import robot_config

        home = robot_config.get_robotframework_ls_home()

        pyexe: Union[bytes, str] = sys.executable
        if isinstance(pyexe, bytes):
            pyexe = pyexe.decode("utf-8", "replace")

        digest: str = get_digest_from_string(pyexe)

        v = cls.get_robot_version()

        # Note: _v1: information on the mtime of the libspec sources now available.
        return os.path.join(home, "specs", cls.INTERNAL_VERSION, "%s_%s" % (digest, v))

    @classmethod
    def get_internal_builtins_libspec_dir(cls, internal_libspec_dir=None):
        return os.path.join(
            internal_libspec_dir or cls.get_internal_libspec_dir(), "builtins"
        )

    # On v2 we disambiguate by using a hash for the filenames if we're generating
    # a libspec for a target filename.
    INTERNAL_VERSION = "v2"

    def __init__(
        self,
        builtin_libspec_dir: Optional[str] = None,
        user_libspec_dir: Optional[str] = None,
        dir_cache_dir: Optional[str] = None,
        observer: Optional[IFSObserver] = None,
        endpoint: Optional[IEndPoint] = None,
        pre_generate_libspecs: bool = False,
        cache_libspec_dir: Optional[str] = None,
    ):
        """
        :param __internal_libspec_dir__:
            Only to be used in tests (to regenerate the builtins)!
        """
        from robocorp_ls_core import watchdog_wrapper
        from robocorp_ls_core.cache import DirCache
        from robotframework_ls import robot_config

        dir_cache = DirCache(
            dir_cache_dir
            or os.path.join(robot_config.get_robotframework_ls_home(), ".cache")
        )

        self._libspec_dir = self.get_internal_libspec_dir()

        self._user_libspec_dir = user_libspec_dir or os.path.join(
            self._libspec_dir, "user"
        )
        self._cache_libspec_dir = cache_libspec_dir or os.path.join(
            self._libspec_dir, "cache"
        )
        self._builtins_libspec_dir = (
            builtin_libspec_dir
            or self.get_internal_builtins_libspec_dir(self._libspec_dir)
        )
        log.info("User libspec dir: %s", self._user_libspec_dir)
        log.info("Builtins libspec dir: %s", self._builtins_libspec_dir)
        log.info("Cache libspec dir: %s", self._cache_libspec_dir)

        try:
            os.makedirs(self._user_libspec_dir)
        except:
            # Ignore exception if it's already created.
            pass
        try:
            os.makedirs(self._builtins_libspec_dir)
        except:
            # Ignore exception if it's already created.
            pass
        try:
            os.makedirs(self._cache_libspec_dir)
        except:
            # Ignore exception if it's already created.
            pass

        self.pre_generate_libspecs = pre_generate_libspecs

        if pre_generate_libspecs:
            from robotframework_ls.impl.libspec_markdown_conversion import (
                LibspecMarkdownConversion,
            )

            self.libspec_markdown_conversion: Optional[
                LibspecMarkdownConversion
            ] = LibspecMarkdownConversion(self)
        else:
            self.libspec_markdown_conversion = None

        self._libspec_warmup = LibspecWarmup(endpoint, dir_cache)

        self._libspec_failures_cache: Dict[
            tuple, str
        ] = {}  # key -> error creating libspec

        self._main_thread = threading.current_thread()

        if observer is None:
            from robocorp_ls_core.watchdog_wrapper import create_observer

            log.info("No observer passed to LibspecManager (creating dummy observer).")

            observer = create_observer("dummy", ())
        self._fs_observer = observer

        self._file_changes_notifier = watchdog_wrapper.create_notifier(
            self._on_file_changed, timeout=0.5, extensions=(".libspec", ".py")
        )

        # Spec info found in the workspace
        self._workspace_folder_uri_to_folder_info: Dict[str, _FolderInfo] = {}
        self._additional_pythonpath_folder_to_folder_info: Dict[str, _FolderInfo] = {}

        # Spec info found in the pythonpath
        pythonpath_folder_to_folder_info: Dict[str, _FolderInfo] = {}
        found = set()
        for path in sys.path:
            if path:
                try:
                    resolved = Path(path).resolve()  # also solves links.
                except:
                    log.exception(
                        f"Unable to Path.resolve({path!r}) (resolving PYTHONPATH entries)."
                    )
                    continue

                if resolved in found:
                    continue
                found.add(resolved)
                if os.path.isdir(path):
                    pythonpath_folder_to_folder_info[path] = _FolderInfo(
                        path, recursive=False
                    )

        self._pythonpath_folder_to_folder_info: Dict[
            str, _FolderInfo
        ] = pythonpath_folder_to_folder_info

        # Spec info found in internal dirs (autogenerated)
        self._internal_folder_to_folder_info: Dict[str, _FolderInfo] = {
            self._user_libspec_dir: _FolderInfo(
                self._user_libspec_dir, recursive=False
            ),
            self._builtins_libspec_dir: _FolderInfo(
                self._builtins_libspec_dir, recursive=False
            ),
        }

        # Must be set from the outside world when needed.
        self.config = None

        if self.pre_generate_libspecs:
            log.debug("Generating builtin libraries libspec.")
            self._libspec_warmup.gen_builtin_libraries(self)

        log.debug("Synchronizing internal caches.")
        self._synchronize()
        log.debug("Finished initializing LibspecManager.")

    def schedule_conversion_to_markdown(self, spec_filename: str):
        if self.libspec_markdown_conversion is not None:
            self.libspec_markdown_conversion.schedule_conversion_to_markdown(
                spec_filename
            )

    @property
    def fs_observer(self) -> IFSObserver:
        return self._fs_observer

    def _check_in_main_thread(self):
        curr_thread = threading.current_thread()
        if self._main_thread is not curr_thread:
            raise AssertionError(
                f"This may only be called at the thread: {self._main_thread}. Current thread: {curr_thread}"
            )

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE,
        )
        from robocorp_ls_core.basic import make_unique

        self._check_in_main_thread()
        from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHONPATH

        self._config = config
        existing_entries = set(self._additional_pythonpath_folder_to_folder_info.keys())
        if config is not None:
            pythonpath_entries = set(
                config.get_setting(OPTION_ROBOT_PYTHONPATH, list, [])
            )
            for new_pythonpath_entry in pythonpath_entries:
                new_pythonpath_entry = os.path.abspath(new_pythonpath_entry)

                if new_pythonpath_entry not in existing_entries:
                    self.add_additional_pythonpath_folder(new_pythonpath_entry)
            for old_entry in existing_entries:
                if old_entry not in pythonpath_entries:
                    self.remove_additional_pythonpath_folder(old_entry)

            pre_generate = []
            pre_generate.extend(
                config.get_setting(OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE, list, [])
            )

            if self.pre_generate_libspecs:
                log.debug("Generating user/pythonpath libraries libspec.")
                self._libspec_warmup.gen_user_libraries(self, make_unique(pre_generate))

    @property
    def user_libspec_dir(self):
        return self._user_libspec_dir

    @property
    def cache_libspec_dir(self):
        return self._cache_libspec_dir

    def _on_file_changed(self, spec_file, folder_info_on_change_spec):
        log.debug("File change detected: %s", spec_file)

        # Check if the cache related to libspec generation failure must be
        # cleared.
        fix = False
        for cache_key in self._libspec_failures_cache:
            libname = cache_key[0]
            if libname in spec_file:
                fix = True
                break
        if fix:
            new = {}
            for cache_key, value in self._libspec_failures_cache.items():
                libname = cache_key[0]
                if libname not in spec_file:
                    new[cache_key] = value
            # Always set as a whole (to avoid racing conditions).
            self._libspec_failures_cache = new

        # Notify _FolderInfo._on_change_spec
        if spec_file.lower().endswith(".libspec"):
            folder_info_on_change_spec(spec_file)

    def add_workspace_folder(self, folder_uri: str):
        self._check_in_main_thread()
        from robocorp_ls_core import uris

        if folder_uri not in self._workspace_folder_uri_to_folder_info:
            log.debug("Added workspace folder: %s", folder_uri)
            cp = self._workspace_folder_uri_to_folder_info.copy()
            folder_info = cp[folder_uri] = _FolderInfo(
                uris.to_fs_path(folder_uri), recursive=True
            )
            self._workspace_folder_uri_to_folder_info = cp
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()
        else:
            log.debug("Workspace folder already added: %s", folder_uri)

    def remove_workspace_folder(self, folder_uri: str):
        self._check_in_main_thread()
        if folder_uri in self._workspace_folder_uri_to_folder_info:
            log.debug("Removed workspace folder: %s", folder_uri)
            cp = self._workspace_folder_uri_to_folder_info.copy()
            folder_info = cp.pop(folder_uri, NULL)
            folder_info.dispose()
            self._workspace_folder_uri_to_folder_info = cp
        else:
            log.debug("Workspace folder already removed: %s", folder_uri)

    def add_additional_pythonpath_folder(self, folder_path):
        self._check_in_main_thread()
        if folder_path not in self._additional_pythonpath_folder_to_folder_info:
            log.debug("Added additional pythonpath folder: %s", folder_path)
            cp = self._additional_pythonpath_folder_to_folder_info.copy()
            folder_info = cp[folder_path] = _FolderInfo(folder_path, recursive=True)
            self._additional_pythonpath_folder_to_folder_info = cp
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()
        else:
            log.debug("Additional pythonpath folder already added: %s", folder_path)

    def remove_additional_pythonpath_folder(self, folder_path):
        self._check_in_main_thread()
        if folder_path in self._additional_pythonpath_folder_to_folder_info:
            log.debug("Removed additional pythonpath folder: %s", folder_path)
            cp = self._additional_pythonpath_folder_to_folder_info.copy()
            folder_info = cp.pop(folder_path, NULL)
            folder_info.dispose()
            self._additional_pythonpath_folder_to_folder_info = cp
        else:
            log.debug("Additional pythonpath folder already removed: %s", folder_path)

    def synchronize_workspace_folders(self):
        for folder_info in self._workspace_folder_uri_to_folder_info.values():
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()

    def synchronize_pythonpath_folders(self):
        for folder_info in self._pythonpath_folder_to_folder_info.values():
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()

    def synchronize_additional_pythonpath_folders(self):
        for folder_info in self._additional_pythonpath_folder_to_folder_info.values():
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()

    def synchronize_internal_libspec_folders(self):
        for folder_info in self._internal_folder_to_folder_info.values():
            folder_info.start_watch(self._fs_observer, self._file_changes_notifier)
            folder_info.synchronize()

    def _synchronize(self):
        """
        Updates the internal caches related to the tracked .libspec files found.

        This can be a slow call as it may traverse the whole workspace folders
        hierarchy, so, it should be used only during startup to fill the initial
        info.
        """
        self.synchronize_workspace_folders()
        self.synchronize_pythonpath_folders()
        self.synchronize_additional_pythonpath_folders()
        self.synchronize_internal_libspec_folders()

    def collect_all_tracked_folders(self) -> Iterator[str]:
        from robocorp_ls_core import uris

        for uri in self._workspace_folder_uri_to_folder_info.keys():
            yield uris.to_fs_path(uri)

        yield from self._pythonpath_folder_to_folder_info.keys()

        yield from self._additional_pythonpath_folder_to_folder_info.keys()

    def iter_lib_info(self, builtin=False):
        """
        :rtype: generator(_LibInfo)
        """
        # Note: the iteration order is important (first ones are visited earlier
        # and have higher priority).
        iter_in = []
        for (_uri, info) in self._workspace_folder_uri_to_folder_info.items():
            if info.libspec_canonical_filename_to_info:
                iter_in.append((info.libspec_canonical_filename_to_info, False))

        for (_uri, info) in self._pythonpath_folder_to_folder_info.items():
            if info.libspec_canonical_filename_to_info:
                iter_in.append((info.libspec_canonical_filename_to_info, False))

        for (_uri, info) in self._additional_pythonpath_folder_to_folder_info.items():
            if info.libspec_canonical_filename_to_info:
                iter_in.append((info.libspec_canonical_filename_to_info, False))

        if builtin:
            info = self._internal_folder_to_folder_info[self._builtins_libspec_dir]
            if info.libspec_canonical_filename_to_info:
                iter_in.append((info.libspec_canonical_filename_to_info, True))
        else:
            for (_uri, info) in self._internal_folder_to_folder_info.items():
                if info.libspec_canonical_filename_to_info:
                    iter_in.append((info.libspec_canonical_filename_to_info, True))

        for canonical_filename_to_info, can_regenerate in iter_in:
            for canonical_spec_filename, info in list(
                canonical_filename_to_info.items()
            ):

                if info is None:
                    info = canonical_filename_to_info[
                        canonical_spec_filename
                    ] = _load_lib_info(self, canonical_spec_filename, can_regenerate)

                # Note: we could end up yielding a library with the same name
                # multiple times due to its scope. It's up to the caller to
                # validate that.
                if info is not None and info.library_doc is not None:
                    yield info

    def get_library_names(self):
        return sorted(
            set(lib_info.library_doc.name for lib_info in self.iter_lib_info())
        )

    def _get_cached_error(
        self,
        libname,
        *,
        is_builtin=False,
        target_file: Optional[str] = None,
        args: Optional[str] = None,
    ) -> Optional[str]:
        cache_key = (libname, is_builtin, target_file, args)
        return self._libspec_failures_cache.get(cache_key)

    def _create_libspec(
        self,
        libname,
        *,
        is_builtin=False,
        target_file: Optional[str] = None,
        args: Optional[str] = None,
    ) -> Optional[str]:
        """
        :param target_file:
            If given this is the library file (i.e.: c:/foo/bar.py) which is the
            actual library we're creating the spec for.
        """
        cache_key = (libname, is_builtin, target_file, args)
        previous = self._libspec_failures_cache.get(cache_key, Sentinel.SENTINEL)
        if previous is not Sentinel.SENTINEL:
            return typing.cast(Optional[str], previous)

        error_creating = self._cached_create_libspec(
            libname, is_builtin, target_file, args
        )
        if error_creating is not None:
            # Always set as a whole (to avoid racing conditions).
            cp = self._libspec_failures_cache.copy()
            cp[cache_key] = error_creating
            self._libspec_failures_cache = cp

        return error_creating

    def _subprocess_check_output(self, *args, **kwargs):
        # Only done for mocking.
        from robocorp_ls_core.subprocess_wrapper import subprocess

        return subprocess.check_output(*args, **kwargs)

    def _cached_create_libspec(
        self,
        libname: str,
        is_builtin: bool,
        target_file: Optional[str],
        args: Optional[str],
        *,
        _internal_force_text=False,  # Should only be set from within this function.
    ) -> Optional[str]:
        """
        Returns an error message if it wasn't able to generate it or None if
        it did generate it.
        """
        from robotframework_ls.impl import robot_constants

        if not is_builtin:
            if not target_file:
                is_builtin = libname in robot_constants.STDLIBS

        import time
        from robocorp_ls_core.subprocess_wrapper import subprocess
        from robocorp_ls_core.robotframework_log import get_log_level

        acquire_mutex = timed_acquire_mutex_for_spec_filename

        if _internal_force_text:
            # In this case this is a recursive call and we already have the lock.
            acquire_mutex = NULL

        log_exception = log.exception
        if is_builtin and libname == "Dialogs" and get_log_level() < 1:
            # Dialogs may have dependencies that are not available, so, don't show
            # it unless verbose mode is enabled.
            log_exception = log.debug

        if not libname.replace(".", "").replace("/", "").replace("\\", "").strip():
            return f"Unable to generate libspec for: {libname}"

        additional_path = None
        additional_path_exists = False

        log_time = True
        cwd = None

        if target_file is not None:
            additional_path = os.path.dirname(target_file)
            if os.path.splitext(os.path.basename(target_file))[0] == "__init__":
                additional_path = os.path.dirname(additional_path)

            additional_path_exists = os.path.exists(additional_path)
            if additional_path and additional_path_exists:
                cwd = additional_path
            if libname.endswith(("/", "\\")):
                libname = libname[:-1]
            libname = os.path.basename(libname)
            if libname.lower().endswith((".py", ".class", ".java")):
                libname = os.path.splitext(libname)[0]

        curtime = time.time()

        try:
            try:
                call = [sys.executable]
                major_version = self.get_robot_major_version()
                if major_version < 4:
                    call.extend("-m robot.libdoc --format XML".split())
                else:
                    call.extend(
                        "-m robot.libdoc --format XML --specdocformat RAW".split()
                    )

                if additional_path and additional_path_exists:
                    call.extend(["-P", os.path.normpath(additional_path)])

                if _internal_force_text:
                    call.append("--docformat")
                    call.append("text")

                # Note: always set as a whole, so, iterate in generator is thread-safe.
                for entry in self._additional_pythonpath_folder_to_folder_info:
                    if os.path.exists(entry):
                        call.extend(["-P", os.path.normpath(entry)])

                if not args:
                    call.append(libname)
                else:
                    call.append("::".join([libname, args]))

                libspec_filename = self._compute_libspec_filename(
                    libname, is_builtin, target_file, args
                )

                log.debug(f"Obtaining mutex to generate libspec: {libspec_filename}.")
                with acquire_mutex(libspec_filename):  # Could fail.
                    log.debug(
                        f"Obtained mutex to generate libspec: {libspec_filename}."
                    )
                    call.append(libspec_filename)

                    mtime: float = -1
                    try:
                        mtime = os.path.getmtime(libspec_filename)
                    except:
                        pass

                    log.debug(
                        "Generating libspec for: %s.\nCwd:%s\nCommand line:\n%s",
                        libname,
                        cwd,
                        " ".join(call),
                    )
                    try:
                        try:
                            # Note: stdout is always subprocess.PIPE in this call.
                            # Note: the env is always inherited (the process which has
                            # the LibspecManager must be the target env already).
                            self._subprocess_check_output(
                                call,
                                stderr=subprocess.STDOUT,
                                stdin=subprocess.PIPE,
                                cwd=cwd,
                            )
                        except OSError as e:
                            log.exception("Error calling: %s", call)
                            # We may have something as: Ignore OSError: [WinError 6] The handle is invalid,
                            # give the result based on whether the file changed on disk.
                            try:
                                if mtime != os.path.getmtime(libspec_filename):
                                    _dump_spec_filename_additional_info(
                                        self,
                                        libspec_filename,
                                        is_builtin=is_builtin,
                                        obtain_mutex=False,
                                    )
                                    return None
                            except:
                                pass

                            log.debug("Not retrying after OSError failure.")
                            return str(e)

                    except subprocess.CalledProcessError as e:
                        if not _internal_force_text:
                            if (
                                b"reST format requires 'docutils' module to be installed"
                                in e.output
                            ):
                                return self._cached_create_libspec(
                                    libname,
                                    is_builtin,
                                    target_file,
                                    args,
                                    _internal_force_text=True,
                                )

                        log_exception(
                            "Error creating libspec: %s.\nReturn code: %s\nOutput:\n%s",
                            libname,
                            e.returncode,
                            e.output,
                        )
                        bytes_output = e.output
                        output = bytes_output.decode("utf-8", "replace")

                        # Remove things we don't want to show.
                        for s in ("Try --help", "--help", "Traceback"):
                            index = output.find(s)
                            if index >= 0:
                                output = output[:index].strip()

                        if output:
                            return output
                        return f"Error creating libspec: {output}"

                    _dump_spec_filename_additional_info(
                        self,
                        libspec_filename,
                        is_builtin=is_builtin,
                        obtain_mutex=False,
                    )
                    return None
            except Exception as e:
                log_exception("Error creating libspec: %s", libname)
                return str(e)
        finally:
            if log_time:
                delta = time.time() - curtime
                log.debug("Took: %.2fs to generate info for: %s" % (delta, libname))

    def dispose(self):
        self._file_changes_notifier.dispose()
        if self.libspec_markdown_conversion is not None:
            self.libspec_markdown_conversion.dispose()

    def _compute_libspec_filename(
        self,
        libname: str,
        is_builtin: bool = False,
        target_file: Optional[str] = None,
        args: Optional[str] = None,
    ):
        from robotframework_ls.impl import robot_constants

        libspec_dir = self._user_libspec_dir
        if libname in robot_constants.STDLIBS:
            libspec_dir = self._builtins_libspec_dir

        if target_file:
            if args:
                digest = (
                    get_digest_from_string(target_file)
                    + "_"
                    + get_digest_from_string(args)
                )
            else:
                digest = get_digest_from_string(target_file)

            libspec_filename = os.path.join(libspec_dir, digest + ".libspec")
        elif not args:
            libspec_filename = os.path.join(libspec_dir, libname + ".libspec")
        else:
            digest = get_digest_from_string(args)
            libspec_filename = os.path.join(libspec_dir, libname + digest + ".libspec")
        return libspec_filename

    def _do_create_libspec_on_get(
        self, libname, target_file: Optional[str], args: Optional[str], is_builtin: bool
    ) -> Optional[str]:
        error_creating = self._create_libspec(
            libname, target_file=target_file, args=args, is_builtin=is_builtin
        )

        if error_creating is None:
            self.synchronize_internal_libspec_folders()
        return error_creating

    def _get_library_target_filename(
        self, libname: str, current_doc_uri: Optional[str] = None
    ) -> Optional[str]:
        from robocorp_ls_core import uris

        target_file: Optional[str] = None
        libname_lower = libname.lower()

        if os.path.isabs(libname):
            target_file = libname
        else:
            # Check if it maps to a file in the filesystem
            if current_doc_uri is not None:
                cwd = os.path.dirname(uris.to_fs_path(current_doc_uri))
                if cwd and os.path.isdir(cwd):
                    f = os.path.join(cwd, libname)
                    if os.path.isdir(f):
                        f = os.path.join(f, "__init__.py")

                    if os.path.exists(f):
                        target_file = f

                    elif not libname_lower.endswith(".py"):
                        f += ".py"
                        if os.path.exists(f):
                            target_file = f

            if target_file is None and libname.endswith((".py", ".class", ".java")):
                iter_in_pythonpath_directories = itertools.chain(
                    self._additional_pythonpath_folder_to_folder_info.keys(),
                    self._pythonpath_folder_to_folder_info.keys(),
                )

                # https://github.com/robocorp/robotframework-lsp/issues/266
                # If the user specifies a file, we don't just search the current
                # relative folder, we also need to search for relative entries
                # in the whole PYTHONPATH.
                for entry in iter_in_pythonpath_directories:
                    check_target = os.path.join(entry, libname)
                    if os.path.exists(check_target):
                        target_file = check_target
                        break

        return target_file

    def get_library_doc_or_error(
        self,
        libname: str,
        create: bool,
        completion_context: ICompletionContext,
        builtin: bool = False,
        args: Optional[str] = None,
    ) -> ILibraryDocOrError:
        """
        :param libname:
            It may be a library name, a relative path to a .py file or an
            absolute path to a .py file.
        """
        from robotframework_ls.impl import robot_constants
        from robotframework_ls.impl import ast_utils

        libname_lower = libname.lower()
        target_file: str = ""
        normalized_target_file: str = ""
        pre_error_msg: str = ""

        config = self.config
        libraries_libdoc_needs_args_lower: Set[str]
        if config is not None:
            libraries_libdoc_needs_args_lower = set(
                str(x).lower()
                for x in config.get_setting(
                    OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS,
                    list,
                    ["remote", "fakerlib"],
                )
            )
        else:
            libraries_libdoc_needs_args_lower = {"remote", "fakerlib"}

        # Note that experimentally, using '*' may pass args to all libraries.
        if (
            libname_lower not in libraries_libdoc_needs_args_lower
            and "*" not in libraries_libdoc_needs_args_lower
        ):
            args = None

        if args:
            if "{" in args:
                # We need to resolve the arguments if there are variables in it.
                from robotframework_ls.impl.variable_resolve import (
                    ResolveVariablesContext,
                )

                assert completion_context.config is config

                args, unresolved = ResolveVariablesContext(
                    completion_context
                ).token_value_and_unresolved_resolving_variables(
                    ast_utils.create_token(args)
                )

                pre_error_msg = (
                    "It was not possible to statically resolve the following variables:\n%s\nFollow-up error:\n"
                    % (", ".join(str(x) for x in unresolved),)
                )

            args = args.replace("\\\\", "\\")

        if not builtin:
            found_target_filename = self._get_library_target_filename(
                libname, completion_context.doc.uri
            )
            if found_target_filename:
                target_file = found_target_filename
                normalized_target_file = normalize_filename(target_file)
            else:
                builtin = libname in robot_constants.STDLIBS

        if libname_lower.endswith((".py", ".class", ".java")):
            libname_lower = os.path.splitext(libname_lower)[0]

        if "/" in libname_lower or "\\" in libname_lower:
            libname_lower = os.path.basename(libname_lower)

        lib_info: _LibInfo
        for lib_info in self.iter_lib_info(builtin=builtin):
            library_doc = lib_info.library_doc

            # If it maps to a file in the filesystem, that's what we need to match,
            # otherwise, match just by its name.
            # Note: this is only valid for the cases where we can regenerate the info
            # for cases where this information is builtin, only match by the name.
            if target_file and lib_info._can_regenerate:
                if args:
                    digest = (
                        get_digest_from_string(target_file)
                        + "_"
                        + get_digest_from_string(args)
                    )
                    found = library_doc.filename.endswith(digest + ".libspec")

                else:
                    found = bool(
                        library_doc.source
                        and normalize_filename(library_doc.source)
                        == normalized_target_file
                    )
                    if not found:
                        try:
                            found = bool(
                                library_doc.source
                                and os.path.samefile(library_doc.source, target_file)
                            )
                        except:
                            # os.path.samefile touches the filesystem, so, it can
                            # raise an exception.
                            found = False
            else:
                if not args:
                    found = bool(
                        library_doc.name and library_doc.name.lower() == libname_lower
                    )
                else:
                    digest = get_digest_from_string(args)
                    found = library_doc.filename.endswith(
                        os.path.normcase(libname + digest + ".libspec")
                    )

            if found:
                if not lib_info.verify_sources_sync():
                    if create:
                        # Found but it's not in sync. Try to regenerate (don't proceed
                        # because we don't want to match a lower priority item, so,
                        # regenerate and get from the cache without creating).
                        self._do_create_libspec_on_get(
                            libname, target_file, args, is_builtin=builtin
                        )

                        # Note: get even if it if was not created (we may match
                        # a lower priority library).
                        return self.get_library_doc_or_error(
                            libname,
                            create=False,
                            completion_context=completion_context,
                            builtin=builtin,
                            args=args,
                        )
                    else:
                        # Not in sync and it should not be created, just skip it.
                        continue
                else:
                    return _LibraryDocOrError(library_doc, None)

        if create:
            error_msg = self._do_create_libspec_on_get(
                libname, target_file, args, is_builtin=builtin
            )
            if error_msg is None:
                return self.get_library_doc_or_error(
                    libname,
                    create=False,
                    completion_context=completion_context,
                    builtin=builtin,
                    args=args,
                )
            return _LibraryDocOrError(None, pre_error_msg + error_msg)

        error_msg = self._get_cached_error(
            libname, is_builtin=builtin, target_file=target_file, args=args
        )
        if error_msg:
            log.debug("Unable to get library named: %s. Reason: %s", libname, error_msg)
            return _LibraryDocOrError(None, pre_error_msg + error_msg)

        msg = f"Unable to find library named: {libname}"
        log.debug(msg)
        return _LibraryDocOrError(None, pre_error_msg + msg)


class _LibraryDocOrError:
    def __init__(self, library_doc: Optional[ILibraryDoc], error: Optional[str]):
        self.library_doc = library_doc
        self.error = error

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ILibraryDocOrError = check_implements(self)



def Method1():
    print('m1')

def Method2():
    print('m2 before')
    Method1()
    print('m2 after')

def Method3():
    print('m3 before')
    Method2()
    print('m3 after')
   
if __name__ == '__main__': 
    Method3()
    print('TEST SUCEEDED!')



from typing import List

OPTION_ROBOT_PYTHON_EXECUTABLE = "robot.python.executable"

OPTION_ROBOT_PYTHON_ENV = "robot.python.env"

OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT = "robot.language-server.tcp-port"

OPTION_ROBOT_LANGUAGE_SERVER_ARGS = "robot.language-server.args"

OPTION_ROBOT_VARIABLES = "robot.variables"
OPTION_ROBOT_VARIABLES_LOAD_FROM_ARGUMENTS_FILE = "robot.loadVariablesFromArgumentsFile"

OPTION_ROBOT_LINT_ROBOCOP_ENABLED = "robot.lint.robocop.enabled"

OPTION_ROBOT_LINT_ENABLED = "robot.lint.enabled"
OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS = "robot.lint.undefinedKeywords"
OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES = "robot.lint.undefinedLibraries"
OPTION_ROBOT_LINT_UNDEFINED_RESOURCES = "robot.lint.undefinedResources"
OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS = "robot.lint.undefinedVariableImports"
OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS = "robot.lint.keywordCallArguments"

OPTION_ROBOT_LINT_VARIABLES = "robot.lint.variables"
OPTION_ROBOT_LINT_IGNORE_VARIABLES = "robot.lint.ignoreVariables"


OPTION_ROBOT_CODE_FORMATTER = "robot.codeFormatter"

OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY = "robotidy"
OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY = "builtinTidy"

OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT = "robot.completions.keywords.format"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_FIRST_UPPER = "First upper"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_TITLE = "Title Case"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_LOWER = "all lower"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_UPPER = "ALL UPPER"


OPTION_ROBOT_PYTHONPATH = "robot.pythonpath"
OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS = "robot.libraries.libdoc.needsArgs"
OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE = "robot.libraries.libdoc.preGenerate"

OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS = (
    "robot.workspaceSymbolsOnlyForOpenDocs"
)

OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM = "robot.completions.section_headers.form"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL = "plural"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_SINGULAR = "singular"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_BOTH = "both"

OPTION_ROBOT_SHOW_CODE_LENSES = "robot.codeLens.enable"

# Options which must be set as environment variables.
ENV_OPTION_ROBOT_DAP_TIMEOUT = "ROBOT_DAP_TIMEOUT"


ALL_ROBOT_OPTIONS = frozenset(
    (
        OPTION_ROBOT_PYTHON_EXECUTABLE,
        OPTION_ROBOT_PYTHON_ENV,
        OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT,
        OPTION_ROBOT_LANGUAGE_SERVER_ARGS,
        OPTION_ROBOT_VARIABLES,
        OPTION_ROBOT_LINT_ROBOCOP_ENABLED,
        OPTION_ROBOT_PYTHONPATH,
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
        OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS,
        OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT,
        OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS,
        OPTION_ROBOT_LINT_ENABLED,
        OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS,
        OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES,
        OPTION_ROBOT_LINT_UNDEFINED_RESOURCES,
        OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
    )
)


CHECK_IF_LIBRARIES_INSTALLED: List[str] = [
    # robotframework-appiumlibrary
    "AppiumLibrary",
    # robotframework-datadriver
    "DataDriver",
    # robotframework-mainframe3270
    "Mainframe3270",
    # robotframework-sshlibrary
    "SSHLibrary",
    # robotframework-whitelibrary
    "WhiteLibrary",
    # rpaframework-google
    "RPA.Cloud.Google",
    # rpaframework
    "RPA.Archive",
    "RPA.Browser",
    "RPA.Browser.Selenium",
    "RPA.Cloud.AWS",
    "RPA.Cloud.Azure",
    "RPA.Crypto",
    "RPA.Database",
    "RPA.Desktop",
    "RPA.Desktop.Clipboard",
    "RPA.Desktop.OperatingSystem",
    "RPA.Desktop.Windows",
    "RPA.Dialogs",
    "RPA.Email.Exchange",
    "RPA.Email.ImapSmtp",
    "RPA.Excel.Application",
    "RPA.Excel.Files",
    "RPA.FileSystem",
    "RPA.FTP",
    "RPA.HTTP",
    "RPA.Images",
    "RPA.JavaAccessBridge",
    "RPA.JSON",
    "RPA.Netsuite",
    "RPA.Notifier",
    "RPA.Outlook.Application",
    "RPA.PDF",
    "RPA.Robocloud.Items",
    "RPA.Robocloud.Secrets",
    "RPA.Robocorp.Process",
    "RPA.Robocorp.Vault",
    "RPA.Robocorp.WorkItems",
    "RPA.RobotLogListener",
    "RPA.Salesforce",
    "RPA.SAP",
    "RPA.Slack",
    "RPA.Tables",
    "RPA.Tasks",
    "RPA.Twitter",
    "RPA.Word.Application",
]



from pathlib import Path
from typing import Tuple, Dict, List, Iterable, Optional, Any, Pattern

import click
import re

from robotidy.app import Robotidy
from robotidy.transformers import load_transformers
from robotidy.files import read_pyproject_config, find_and_read_config, DEFAULT_EXCLUDES
from robotidy.utils import (
    GlobalFormattingConfig,
    split_args_from_name_or_path,
    remove_rst_formatting,
    RecommendationFinder,
)
from robotidy.decorators import catch_exceptions
from robotidy.version import __version__


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
HELP_MSG = f"""
Version: {__version__}

Robotidy is a tool for formatting Robot Framework source code.
See examples at the end of this help message too see how you can use Robotidy.
Full documentation available at https://robotidy.readthedocs.io .
"""
EPILOG = """
Examples:
  # Format `path/to/src.robot` file
  $ robotidy path/to/src.robot

  # Format every Robot Framework file inside `dir_name` directory
  $ robotidy dir_name

  # List available transformers:
  $ robotidy --list
  
  # Display transformer documentation
  $ robotidy --desc <TRANSFORMER_NAME>

  # Format `src.robot` file using `SplitTooLongLine` transformer only
  $ robotidy --transform SplitTooLongLine src.robot

  # Format `src.robot` file using `SplitTooLongLine` transformer only and configured line length 140
  $ robotidy --transform SplitTooLongLine:line_length=140 src.robot

"""


class RawHelp(click.Command):
    def format_help_text(self, ctx, formatter):
        if self.help:
            formatter.write_paragraph()
            for line in self.help.split("\n"):
                formatter.write_text(line)

    def format_epilog(self, ctx, formatter):
        if self.epilog:
            formatter.write_paragraph()
            for line in self.epilog.split("\n"):
                formatter.write_text(line)


class TransformType(click.ParamType):
    name = "transform"

    def convert(self, value, param, ctx):
        name, args = split_args_from_name_or_path(value.replace(" ", ""))
        return name, args


def parse_opt(opt):
    while opt and opt[0] == "-":
        opt = opt[1:]
    return opt.replace("-", "_")


def validate_config_options(params, config):
    if params is None:
        return
    allowed = {parse_opt(opt) for param in params for opt in param.opts}
    for conf in config:
        if conf not in allowed:
            rec_finder = RecommendationFinder()
            similar = rec_finder.find(conf, list(allowed))
            raise click.NoSuchOption(conf, possibilities=similar)


def read_config(ctx: click.Context, param: click.Parameter, value: Optional[str]) -> Optional[str]:
    # if --config was not used, try to find pyproject.toml or robotidy.toml file
    if value:
        config = read_pyproject_config(value)
    else:
        config = find_and_read_config(ctx.params["src"] or (str(Path(".").resolve()),))
    if not config:
        return
    # Sanitize the values to be Click friendly. For more information please see:
    # https://github.com/psf/black/issues/1458
    # https://github.com/pallets/click/issues/1567
    config = {k: str(v) if not isinstance(v, (list, dict)) else v for k, v in config.items()}
    if "src" in config:
        config["src"] = tuple(config["src"])
    validate_config_options(ctx.command.params, config)
    default_map: Dict[str, Any] = {}
    if ctx.default_map:
        default_map.update(ctx.default_map)
    default_map.update(config)
    ctx.default_map = default_map


def validate_regex_callback(
    ctx: click.Context,
    param: click.Parameter,
    value: Optional[str],
) -> Optional[Pattern]:
    return validate_regex(value)


def validate_regex(value: Optional[str]) -> Optional[Pattern]:
    try:
        return re.compile(value) if value is not None else None
    except re.error:
        raise click.BadParameter("Not a valid regular expression")


def print_description(name: str):
    transformers = load_transformers(None, {}, allow_disabled=True)
    transformer_by_names = {transformer.__class__.__name__: transformer for transformer in transformers}
    if name == "all":
        for tr_name, transformer in transformer_by_names.items():
            click.echo(f"Transformer {tr_name}:")
            click.echo(remove_rst_formatting(transformer.__doc__))
    elif name in transformer_by_names:
        click.echo(f"Transformer {name}:")
        click.echo(remove_rst_formatting(transformer_by_names[name].__doc__))
    else:
        rec_finder = RecommendationFinder()
        similar = rec_finder.find_similar(name, transformer_by_names.keys())
        click.echo(f"Transformer with the name '{name}' does not exist.{similar}")
        return 1
    return 0


def print_transformers_list():
    transformers = load_transformers(None, {}, allow_disabled=True)
    click.echo(
        "To see detailed docs run --desc <transformer_name> or --desc all. "
        "Transformers with (disabled) tag \nare executed only when selected explicitly with --transform or "
        "configured with param `enabled=True`.\n"
        "Available transformers:\n"
    )
    transformer_names = []
    for transformer in transformers:
        disabled = " (disabled)" if not getattr(transformer, "ENABLED", True) else ""
        transformer_names.append(transformer.__class__.__name__ + disabled)
    for name in sorted(transformer_names):
        click.echo(name)


@click.command(cls=RawHelp, help=HELP_MSG, epilog=EPILOG, context_settings=CONTEXT_SETTINGS)
@click.option(
    "--transform",
    "-t",
    type=TransformType(),
    multiple=True,
    metavar="TRANSFORMER_NAME",
    help="Transform files from [PATH(S)] with given transformer",
)
@click.option(
    "--configure",
    "-c",
    type=TransformType(),
    multiple=True,
    metavar="TRANSFORMER_NAME:PARAM=VALUE",
    help="Configure transformers",
)
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True, allow_dash=True),
    is_eager=True,
    metavar="[PATH(S)]",
)
@click.option(
    "--exclude",
    type=str,
    callback=validate_regex_callback,
    help=(
        "A regular expression that matches files and directories that should be"
        " excluded on recursive searches. An empty value means no paths are excluded."
        " Use forward slashes for directories on all platforms."
        f" [default: '{DEFAULT_EXCLUDES}']"
    ),
    show_default=False,
)
@click.option(
    "--extend-exclude",
    type=str,
    callback=validate_regex_callback,
    help=(
        "Like --exclude, but adds additional files and directories on top of the"
        " excluded ones. (Useful if you simply want to add to the default)"
    ),
)
@click.option(
    "--config",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
        path_type=str,
    ),
    is_eager=True,
    callback=read_config,
    help="Read configuration from FILE path.",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=None,
    help="Write changes back to file",
    show_default=True,
)
@click.option(
    "--diff",
    is_flag=True,
    help="Output diff of each processed file.",
    show_default=True,
)
@click.option(
    "--check",
    is_flag=True,
    help="Don't overwrite files and just return status. Return code 0 means nothing would change. "
    "Return code 1 means that at least 1 file would change. Any internal error will overwrite this status.",
    show_default=True,
)
@click.option(
    "-s",
    "--spacecount",
    type=click.types.INT,
    default=4,
    help="The number of spaces between cells in the plain text format.\n",
    show_default=True,
)
@click.option(
    "-ls",
    "--lineseparator",
    type=click.types.Choice(["native", "windows", "unix", "auto"]),
    default="native",
    help="Line separator to use in outputs.\n"
    "native:  use operating system's native line endings\n"
    "windows: use Windows line endings (CRLF)\n"
    "unix:    use Unix line endings (LF)\n"
    "auto:    maintain existing line endings (uses what's used in the first line)",
    show_default=True,
)
@click.option(
    "--separator",
    type=click.types.Choice(["space", "tab"]),
    default="space",
    help="Token separator to use in outputs.\n"
    "space:   use --spacecount spaces to separate tokens\n"
    "tab:     use a single tabulation to separate tokens\n",
    show_default=True,
)
@click.option(
    "-sl",
    "--startline",
    default=None,
    type=int,
    help="Limit robotidy only to selected area. If --endline is not provided, format text only at --startline. "
    "Line numbers start from 1.",
    show_default=True,
)
@click.option(
    "-el",
    "--endline",
    default=None,
    type=int,
    help="Limit robotidy only to selected area. " "Line numbers start from 1.",
    show_default=True,
)
@click.option(
    "--line-length",
    default=120,
    type=int,
    help="Max allowed characters per line",
    show_default=True,
)
@click.option(
    "--list",
    "-l",
    is_eager=True,
    is_flag=True,
    help="List available transformers and exit.",
)
@click.option(
    "--desc",
    "-d",
    default=None,
    metavar="TRANSFORMER_NAME",
    help="Show documentation for selected transformer.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, allow_dash=False),
    default=None,
    metavar="PATH",
    help="Path to output file where source file will be saved",
)
@click.option("-v", "--verbose", is_flag=True, help="More verbose output", show_default=True)
@click.option(
    "--force-order",
    is_flag=True,
    help="Transform files using transformers in order provided in cli",
)
@click.version_option(version=__version__, prog_name="robotidy")
@click.pass_context
@catch_exceptions
def cli(
    ctx: click.Context,
    transform: List[Tuple[str, List]],
    configure: List[Tuple[str, List]],
    src: Tuple[str, ...],
    exclude: Optional[Pattern],
    extend_exclude: Optional[Pattern],
    overwrite: bool,
    diff: bool,
    check: bool,
    spacecount: int,
    lineseparator: str,
    verbose: bool,
    config: Optional[str],
    separator: Optional[str],
    startline: Optional[int],
    endline: Optional[int],
    line_length: int,
    list: bool,
    desc: Optional[str],
    output: Optional[Path],
    force_order: bool,
):
    if list:
        print_transformers_list()
        ctx.exit(0)
    if desc is not None:
        return_code = print_description(desc)
        ctx.exit(return_code)
    if not src:
        if ctx.default_map is not None:
            src = ctx.default_map.get("src", None)
        if not src:
            print("No source path provided. Run robotidy --help to see how to use robotidy")
            ctx.exit(1)

    if exclude is None:
        exclude = re.compile(DEFAULT_EXCLUDES)

    if config and verbose:
        click.echo(f"Loaded {config} configuration file")

    if overwrite is None:
        # None is default, with check not set -> overwrite, with check set -> overwrite only when overwrite flag is set
        overwrite = not check

    formatting_config = GlobalFormattingConfig(
        space_count=spacecount,
        line_sep=lineseparator,
        start_line=startline,
        separator=separator,
        end_line=endline,
        line_length=line_length,
    )
    tidy = Robotidy(
        transformers=transform,
        transformers_config=configure,
        src=src,
        exclude=exclude,
        extend_exclude=extend_exclude,
        overwrite=overwrite,
        show_diff=diff,
        formatting_config=formatting_config,
        verbose=verbose,
        check=check,
        output=output,
        force_order=force_order,
    )
    status = tidy.transform_files()
    ctx.exit(status)



import threading, atexit, sys
import time

try:
    from thread import start_new_thread
except:
    from _thread import start_new_thread


class MyError(Exception):

    def __init__(self, msg):
        return Exception.__init__(self)


def _atexit():
    print('TEST SUCEEDED')
    sys.stderr.write('TEST SUCEEDED\n')
    sys.stderr.flush()
    sys.stdout.flush()


# Register the TEST SUCEEDED msg to the exit of the process.
atexit.register(_atexit)


def thread_func():
    raise MyError('in thread 1')


start_new_thread(thread_func, ())

# Wait for the first to be handled... otherwise, tests can become flaky if
# both stop at the same time only 1 notification may be given for both, whereas
# the test expects 2 notifications.
time.sleep(.5)


def thread_func2(n):
    raise MyError('in thread 2')


th = threading.Thread(target=lambda: thread_func2(1))
th.daemon = True
th.start()

th.join()

# This is a bit tricky: although we waited on the event, there's a slight chance
# that we didn't get the notification because the thread could've stopped executing,
# so, sleep a bit so that the test does not become flaky.
time.sleep(.5)
raise MyError('in main')



if __name__ == '__main__':
    print('Run as main: %s' % (__file__,))
    import sys
    sys.stdout.flush()
    import pydevd
    # Just check that we're already connected
    assert pydevd.GetGlobalDebugger() is not None
    print('finish')
    sys.stdout.flush()
    print('TEST SUCEEDED!')
    sys.stdout.flush()


'''
Entry point module to run a file in the interactive console.
'''
import os
import sys
import traceback
from pydevconsole import InterpreterInterface, process_exec_queue, start_console_server, init_mpl_in_console
from _pydev_bundle._pydev_saved_modules import threading, _queue

from _pydev_bundle import pydev_imports
from _pydevd_bundle.pydevd_utils import save_main_module
from _pydev_bundle.pydev_console_utils import StdIn
from pydevd_file_utils import get_fullname


def run_file(file, globals=None, locals=None, is_module=False):
    module_name = None
    entry_point_fn = None
    if is_module:
        file, _,  entry_point_fn = file.partition(':')
        module_name = file
        filename = get_fullname(file)
        if filename is None:
            sys.stderr.write("No module named %s\n" % file)
            return
        else:
            file = filename

    if os.path.isdir(file):
        new_target = os.path.join(file, '__main__.py')
        if os.path.isfile(new_target):
            file = new_target

    if globals is None:
        m = save_main_module(file, 'pydev_run_in_console')

        globals = m.__dict__
        try:
            globals['__builtins__'] = __builtins__
        except NameError:
            pass  # Not there on Jython...

    if locals is None:
        locals = globals

    if not is_module:
        sys.path.insert(0, os.path.split(file)[0])

    print('Running %s' % file)
    try:
        if not is_module:
            pydev_imports.execfile(file, globals, locals)  # execute the script
        else:
            # treat ':' as a seperator between module and entry point function
            # if there is no entry point we run we same as with -m switch. Otherwise we perform
            # an import and execute the entry point
            if entry_point_fn:
                mod = __import__(module_name, level=0, fromlist=[entry_point_fn], globals=globals, locals=locals)
                func = getattr(mod, entry_point_fn)
                func()
            else:
                # Run with the -m switch
                import runpy
                if hasattr(runpy, '_run_module_as_main'):
                    runpy._run_module_as_main(module_name)
                else:
                    runpy.run_module(module_name)
    except:
        traceback.print_exc()

    return globals


def skip_successful_exit(*args):
    """ System exit in file shouldn't kill interpreter (i.e. in `timeit`)"""
    if len(args) == 1 and args[0] in (0, None):
        pass
    else:
        raise SystemExit(*args)


def process_args(argv):
    setup_args = {'file': '', 'module': False}

    setup_args['port'] = argv[1]
    del argv[1]
    setup_args['client_port'] = argv[1]
    del argv[1]

    module_flag = "--module"
    if module_flag in argv:
        i = argv.index(module_flag)
        if i != -1:
            setup_args['module'] = True
            setup_args['file'] = argv[i + 1]
            del sys.argv[i]
    else:
        setup_args['file'] = argv[1]

    del argv[0]

    return setup_args


#=======================================================================================================================
# main
#=======================================================================================================================
if __name__ == '__main__':
    setup = process_args(sys.argv)

    port = setup['port']
    client_port = setup['client_port']
    file = setup['file']
    is_module = setup['module']

    from _pydev_bundle import pydev_localhost

    if int(port) == 0 and int(client_port) == 0:
        (h, p) = pydev_localhost.get_socket_name()
        client_port = p

    host = pydev_localhost.get_localhost()

    #replace exit (see comments on method)
    #note that this does not work in jython!!! (sys method can't be replaced).
    sys.exit = skip_successful_exit

    connect_status_queue = _queue.Queue()
    interpreter = InterpreterInterface(host, int(client_port), threading.current_thread(), connect_status_queue=connect_status_queue)

    server_thread = threading.Thread(target=start_console_server,
                                     name='ServerThread',
                                     args=(host, int(port), interpreter))
    server_thread.daemon = True
    server_thread.start()

    sys.stdin = StdIn(interpreter, host, client_port, sys.stdin)

    init_mpl_in_console(interpreter)

    try:
        success = connect_status_queue.get(True, 60)
        if not success:
            raise ValueError()
    except:
        sys.stderr.write("Console server didn't start\n")
        sys.stderr.flush()
        sys.exit(1)

    globals = run_file(file, None, None, is_module)

    interpreter.get_namespace().update(globals)

    interpreter.ShowConsole()

    process_exec_queue(interpreter)



'''
@author Fabio Zadrozny
'''
from _pydev_bundle import _pydev_imports_tipper
import inspect
import pytest
import sys
import unittest

try:
    import __builtin__  # @UnusedImport
    BUILTIN_MOD = '__builtin__'
except ImportError:
    BUILTIN_MOD = 'builtins'

IS_JYTHON = sys.platform.find('java') != -1

HAS_WX = False


@pytest.mark.skipif(IS_JYTHON, reason='CPython related test')
class TestCPython(unittest.TestCase):

    def p(self, t):
        for a in t:
            sys.stdout.write('%s\n' % (a,))

    def test_imports3(self):
        tip = _pydev_imports_tipper.generate_tip('os')
        ret = self.assert_in('path', tip)
        self.assertEqual('', ret[2])

    def test_imports2(self):
        try:
            tip = _pydev_imports_tipper.generate_tip('OpenGL.GLUT')
            self.assert_in('glutDisplayFunc', tip)
            self.assert_in('glutInitDisplayMode', tip)
        except ImportError:
            pass

    def test_imports4(self):
        try:
            tip = _pydev_imports_tipper.generate_tip('mx.DateTime.mxDateTime.mxDateTime')
            self.assert_in('now', tip)
        except ImportError:
            pass

    def test_imports5(self):
        tip = _pydev_imports_tipper.generate_tip('%s.list' % BUILTIN_MOD)
        s = self.assert_in('sort', tip)
        self.check_args(
            s,
            '(cmp=None, key=None, reverse=False)',
            '(self, object cmp, object key, bool reverse)',
            '(self, cmp: object, key: object, reverse: bool)',
            '(key=None, reverse=False)',
            '(self, key=None, reverse=False)',
            '(self, cmp, key, reverse)',
            '(self, key, reverse)',
        )

    def test_imports2a(self):
        tips = _pydev_imports_tipper.generate_tip('%s.RuntimeError' % BUILTIN_MOD)
        self.assert_in('__doc__', tips)

    def test_imports2b(self):
        try:
            file
        except:
            pass
        else:
            tips = _pydev_imports_tipper.generate_tip('%s' % BUILTIN_MOD)
            t = self.assert_in('file' , tips)
            self.assertTrue('->' in t[1].strip() or 'file' in t[1])

    def test_imports2c(self):
        try:
            file  # file is not available on py 3
        except:
            pass
        else:
            tips = _pydev_imports_tipper.generate_tip('%s.file' % BUILTIN_MOD)
            t = self.assert_in('readlines' , tips)
            self.assertTrue('->' in t[1] or 'sizehint' in t[1])

    def test_imports(self):
        '''
        You can print_ the results to check...
        '''
        if HAS_WX:
            tip = _pydev_imports_tipper.generate_tip('wxPython.wx')
            self.assert_in('wxApp'        , tip)

            tip = _pydev_imports_tipper.generate_tip('wxPython.wx.wxApp')

            try:
                tip = _pydev_imports_tipper.generate_tip('qt')
                self.assert_in('QWidget'        , tip)
                self.assert_in('QDialog'        , tip)

                tip = _pydev_imports_tipper.generate_tip('qt.QWidget')
                self.assert_in('rect'           , tip)
                self.assert_in('rect'           , tip)
                self.assert_in('AltButton'      , tip)

                tip = _pydev_imports_tipper.generate_tip('qt.QWidget.AltButton')
                self.assert_in('__xor__'      , tip)

                tip = _pydev_imports_tipper.generate_tip('qt.QWidget.AltButton.__xor__')
                self.assert_in('__class__'      , tip)
            except ImportError:
                pass

        tip = _pydev_imports_tipper.generate_tip(BUILTIN_MOD)
#        for t in tip[1]:
#            print_ t
        self.assert_in('object'         , tip)
        self.assert_in('tuple'          , tip)
        self.assert_in('list'          , tip)
        self.assert_in('RuntimeError'   , tip)
        self.assert_in('RuntimeWarning' , tip)

        # Remove cmp as it's not available on py 3
        # t = self.assert_in('cmp' , tip)
        # self.check_args(t, '(x, y)', '(object x, object y)', '(x: object, y: object)') #args

        t = self.assert_in('isinstance' , tip)
        self.check_args(
            t,
            '(object, class_or_type_or_tuple)',
            '(object o, type typeinfo)',
            '(o: object, typeinfo: type)',
            '(obj, class_or_tuple)',
            '(obj, klass_or_tuple)',
        )  # args

        t = self.assert_in('compile' , tip)
        self.check_args(
            t,
            '(source, filename, mode)',
            '()',
            '(o: object, name: str, val: object)',
            '(source, filename, mode, flags, dont_inherit, optimize)',
            '(source, filename, mode, flags, dont_inherit)',
            '(source, filename, mode, flags, dont_inherit, optimize, _feature_version=-1)'
        )  # args

        t = self.assert_in('setattr' , tip)
        self.check_args(
            t,
            '(object, name, value)',
            '(object o, str name, object val)',
            '(o: object, name: str, val: object)',
            '(obj, name, value)',
            '(object, name, val)',
        )  # args

        try:
            import compiler
            compiler_module = 'compiler'
        except ImportError:
            try:
                import ast
                compiler_module = 'ast'
            except ImportError:
                compiler_module = None

        if compiler_module is not None:  # Not available in iron python
            tip = _pydev_imports_tipper.generate_tip(compiler_module)
            if compiler_module == 'compiler':
                self.assert_args('parse', '(buf, mode)', tip)
                self.assert_args('walk', '(tree, visitor, walker, verbose)', tip)
                self.assert_in('parseFile'      , tip)
            else:
                self.assert_args('parse', [
                        '(source, filename, mode)',
                        '(source, filename, mode, type_comments=False, feature_version=None)'
                    ], tip
                )
                self.assert_args('walk', '(node)', tip)
            self.assert_in('parse'          , tip)

    def check_args(self, t, *expected):
        for x in expected:
            if x == t[2]:
                return
        self.fail('Found: %s. Expected: %s' % (t[2], expected))

    def assert_args(self, tok, args, tips):
        if not isinstance(args, (list, tuple)):
            args = (args,)

        for a in tips[1]:
            if tok == a[0]:
                for arg in args:
                    if arg == a[2]:
                        return
                raise AssertionError('%s not in %s', a[2], args)

        raise AssertionError('%s not in %s', tok, tips)

    def assert_in(self, tok, tips):
        for a in tips[1]:
            if tok == a[0]:
                return a
        raise AssertionError('%s not in %s' % (tok, tips))

    def test_search(self):
        s = _pydev_imports_tipper.search_definition('inspect.ismodule')
        (f, line, col), foundAs = s
        self.assertTrue(line > 0)

    def test_dot_net_libraries(self):
        if sys.platform == 'cli':
            tip = _pydev_imports_tipper.generate_tip('System.Drawing')
            self.assert_in('Brushes' , tip)

            tip = _pydev_imports_tipper.generate_tip('System.Drawing.Brushes')
            self.assert_in('Aqua' , tip)

    def test_tips_hasattr_failure(self):

        class MyClass(object):

            def __getattribute__(self, attr):
                raise RuntimeError()

        obj = MyClass()

        _pydev_imports_tipper.generate_imports_tip_for_module(obj)

    def test_inspect(self):

        class C(object):

            def metA(self, a, b):
                pass

        obj = C.metA
        if inspect.ismethod (obj):
            pass
#            print_ obj.im_func
#            print_ inspect.getargspec(obj.im_func)



if __name__ == '__main__':
    import os
    import sys
    port = int(sys.argv[1])
    root_dirname = os.path.dirname(os.path.dirname(__file__))

    if root_dirname not in sys.path:
        sys.path.append(root_dirname)

    del sys.argv

    import pydevd
    print('before pydevd.settrace')
    pydevd.settrace(port=port)
    print('after pydevd.settrace')
    print('TEST SUCEEDED!')




import pydevd
# Some hackery to get the PyDevJsonCommandProcessor which is not exposed.
try:
    json_command_processor = pydevd.get_global_debugger().reader.process_net_command_json.__self__
except:
    json_command_processor = pydevd.get_global_debugger().reader.process_net_command_json.im_self

print(json_command_processor._options.to_json())

print('TEST SUCEEDED!')



"""
Miscellaneous checkers
"""
from pathlib import Path

from robot.api import Token
from robot.parsing.model.blocks import TestCaseSection
from robot.parsing.model.statements import KeywordCall, Return

try:
    from robot.api.parsing import Variable, Comment, EmptyLine, If
except ImportError:
    from robot.parsing.model.statements import Variable, Comment, EmptyLine
try:
    from robot.api.parsing import ReturnStatement, InlineIfHeader, Break, Continue
except ImportError:
    ReturnStatement, InlineIfHeader, Break, Continue = None, None, None, None
from robot.libraries import STDLIBS

from robocop.checkers import VisitorChecker
from robocop.rules import Rule, RuleParam, RuleSeverity
from robocop.utils import (
    ROBOT_VERSION,
    AssignmentTypeDetector,
    normalize_robot_name,
    parse_assignment_sign_type,
    keyword_col,
    token_col,
    get_errors,
)


rules = {
    "0901": Rule(
        rule_id="0901",
        name="keyword-after-return",
        msg="{{ error_msg }}",
        severity=RuleSeverity.WARNING,
        docs="""
        To improve readability use `[Return]` setting at the end of the keyword. If you want to return immediately from 
        the keyword use `RETURN` statement instead (`[Return]` does not return until all steps in the 
        keyword are completed).
        
        Bad::
        
            Keyword
                Step
                [Return]    ${variable}
                ${variable}    Other Step
        
        Good::
        
            Keyword
                Step
                ${variable}    Other Step
                [Return]    ${variable}

        """,
    ),
    "0903": Rule(
        rule_id="0903",
        name="empty-return",
        msg="[Return] is empty",
        severity=RuleSeverity.WARNING,
        docs="""
        `[Return]` statement is used to define variables returned from keyword. If you don't return anything from 
        keyword,  don't use `[Return]`.
        """,
    ),
    "0907": Rule(
        rule_id="0907",
        name="nested-for-loop",
        msg="Nested for loops are not supported. You can use keyword with for loop instead",
        severity=RuleSeverity.ERROR,
        version="<4.0",
        docs="""
        Older versions of Robot Framework did not support nested for loops::
        
            FOR    ${var}    IN RANGE    10
                FOR   ${other_var}   IN    a  b
                    # Nesting supported from Robot Framework 4.0+
                END
            END

        """,
    ),
    "0908": Rule(
        rule_id="0908",
        name="if-can-be-used",
        msg="'{{ run_keyword }}' can be replaced with IF block since Robot Framework 4.0",
        severity=RuleSeverity.INFO,
        version="==4.*",
        docs="""
        Starting from Robot Framework 4.0 `Run Keyword If` and `Run Keyword Unless` can be replaced by IF block.
        """,
    ),
    "0909": Rule(
        RuleParam(
            name="assignment_sign_type",
            default="autodetect",
            converter=parse_assignment_sign_type,
            desc="possible values: 'autodetect' (default), 'none' (''), "
            "'equal_sign' ('=') or space_and_equal_sign (' =')",
        ),
        rule_id="0909",
        name="inconsistent-assignment",
        msg="The assignment sign is not consistent within the file. Expected '{{ expected_sign }}' "
        "but got '{{ actual_sign }}' instead",
        severity=RuleSeverity.WARNING,
        docs="""
        Use only one type of assignment sign in a file. 
        
        Example of rule violation::
        
            *** Keywords ***
            Keyword
                ${var} =  Other Keyword
                No Operation
            
            Keyword 2
                No Operation
                ${var}  ${var2}  Some Keyword  # this assignment doesn't use equal sign while the previous one uses ` =`
        
        By default Robocop looks for most popular assignment sign in the file. It is possible to define expected 
        assignment sign by running::
        
            robocop --configure inconsistent-assignment:assignment_sign_type:equal_sign
        
        You can choose between following signs: 'autodetect' (default), 'none' (''), 'equal_sign' ('=') or 
        space_and_equal_sign (' =').
    
        """,
    ),
    "0910": Rule(
        RuleParam(
            name="assignment_sign_type",
            default="autodetect",
            converter=parse_assignment_sign_type,
            desc="possible values: 'autodetect' (default), 'none' (''), "
            "'equal_sign' ('=') or space_and_equal_sign (' =')",
        ),
        rule_id="0910",
        name="inconsistent-assignment-in-variables",
        msg="The assignment sign is not consistent inside the variables section. Expected '{{ expected_sign }}' "
        "but got '{{ actual_sign }}' instead",
        severity=RuleSeverity.WARNING,
        docs="""
        Use one type of assignment sign in Variables section. 
        
        Example of rule violation::
        
            *** Variables ***
            ${var} =    1
            ${var2}=    2
            ${var3} =   3
            ${var4}     a
            ${var5}     b
        
        By default Robocop looks for most popular assignment sign in the file. It is possible to define expected 
        assignment sign by running::
        
            robocop --configure inconsistent-assignment-in-variables:assignment_sign_type:equal_sign
        
        You can choose between following signs: 'autodetect' (default), 'none' (''), 'equal_sign' ('=') or 
        space_and_equal_sign (' =').
        
        """,
    ),
    "0911": Rule(
        rule_id="0911",
        name="wrong-import-order",
        msg="BuiltIn library import '{{ builtin_import }}' should be placed before '{{ custom_import }}'",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
            *** Settings ***
            Library    Collections
            Library    CustomLibrary
            Library    OperatingSystem  # BuiltIn library defined after custom CustomLibrary

        """,
    ),
    "0912": Rule(
        rule_id="0912",
        name="empty-variable",
        msg="Use built-in variable ${EMPTY} instead of leaving variable without value or using backslash",
        severity=RuleSeverity.INFO,
        docs="""
        Example of rule violation::
        
            *** Variables ***
            ${VAR_NO_VALUE}                   # missing value
            ${VAR_WITH_EMPTY}       ${EMPTY}
            @{MULTILINE_FIRST_EMPTY}
            ...                               # missing value
            ...  value
            ${EMPTY_WITH_BACKSLASH}  \        # used backslash

        """,
    ),
    "0913": Rule(
        rule_id="0913",
        name="can-be-resource-file",
        msg="No tests in '{{ file_name }}' file, consider renaming to '{{ file_name_stem }}.resource'",
        severity=RuleSeverity.INFO,
        docs="""
        If the Robot file contains only keywords or variables it's a good practice to use `.resource` extension.
        """,
    ),
    "0914": Rule(
        rule_id="0914",
        name="if-can-be-merged",
        msg="IF statement can be merged with previous IF (defined in line {{ line }})",
        severity=RuleSeverity.INFO,
        version=">=4.0",
        docs="""
        IF statement follows another IF with identical conditions. It can be possibly merged into one.
        
        Example of rule violation::
        
            IF  ${var} == 4
                Keyword
            END
            # comments are ignored
            IF  ${var}  == 4
                Keyword 2
            END
        
        IF statement is considered identical only if all branches have identical conditions. 
        
        Similar but not identical IF::
        
            IF  ${variable}
                Keyword
            ELSE
                Other Keyword
            END
            IF  ${variable}
                Keyword
            END

        """,
    ),
    "0915": Rule(
        rule_id="0915",
        name="statement-outside-loop",
        msg="{{ name }} {{ statement_type }} used outside a loop",
        severity=RuleSeverity.ERROR,
        version=">=5.0",
        docs="""
        Following keywords and statements should only be used inside loop (``WHILE`` or ``FOR``):
            - ``Exit For Loop``,
            - ``Exit For Loop If``,
            - ``Continue For Loop``,
            - ``Continue For Loop If ``
            - ``CONTINUE``,
            - ``BREAK``
        
        """,
    ),
    "0916": Rule(
        RuleParam(
            name="max_width",
            default=80,
            converter=int,
            desc="Maximum width of IF (in characters) below which it will be recommended to use inline IF",
        ),
        rule_id="0916",
        name="inline-if-can-be-used",
        msg="IF can be replaced with inline IF",
        severity=RuleSeverity.INFO,
        version=">=5.0",
        docs="""
        Short and simple IFs can be replaced with inline IF.
        
        Following IF::
        
            IF    $condition
                BREAK
            END
        
        can be replaced with::
        
            IF    $condition    BREAK

        """,
    ),
}


class ReturnChecker(VisitorChecker):
    """Checker for [Return] and Return From Keyword violations."""

    reports = (
        "keyword-after-return",
        "empty-return",
    )

    def visit_Keyword(self, node):  # noqa
        return_setting_node = None
        keyword_after_return = False
        return_from = False
        error = ""
        for child in node.body:
            if isinstance(child, Return):
                return_setting_node = child
                error = (
                    "[Return] is not defined at the end of keyword. "
                    "Note that [Return] does not quit from keyword but only set variables to be returned"
                )
                if not child.values:
                    self.report("empty-return", node=child, col=child.end_col_offset)
            elif ReturnStatement and isinstance(child, ReturnStatement):  # type: ignore[arg-type]
                return_setting_node = child
                error = "RETURN is not defined at the end of keyword"
            elif not isinstance(child, (EmptyLine, Comment)):
                if return_setting_node is not None:
                    keyword_after_return = True

            if isinstance(child, KeywordCall):
                if return_from:
                    keyword_after_return = True
                    return_setting_node = child
                    error = "Keyword call after 'Return From Keyword'"
                elif normalize_robot_name(child.keyword, remove_prefix="builtin.") == "returnfromkeyword":
                    return_from = True
        if keyword_after_return:
            token = return_setting_node.data_tokens[0]
            self.report("keyword-after-return", error_msg=error, node=token, col=token.col_offset + 1)
        self.generic_visit(node)

    visit_If = visit_For = visit_While = visit_Try = visit_Keyword


class NestedForLoopsChecker(VisitorChecker):
    """Checker for not supported nested FOR loops.

    Deprecated in RF 4.0
    """

    reports = ("nested-for-loop",)

    def visit_ForLoop(self, node):  # noqa
        # For RF 4.0 node is "For" but we purposely don't visit it because nested for loop is allowed in 4.0
        for child in node.body:
            if child.type == "FOR":
                self.report("nested-for-loop", node=child)


class IfBlockCanBeUsed(VisitorChecker):
    """Checker for potential IF block usage in Robot Framework 4.0

    Run Keyword variants (Run Keyword If, Run Keyword Unless) can be replaced with IF in RF 4.0
    """

    reports = ("if-can-be-used",)
    run_keyword_variants = {"runkeywordif", "runkeywordunless"}

    def visit_KeywordCall(self, node):  # noqa
        if not node.keyword:
            return
        if normalize_robot_name(node.keyword, remove_prefix="builtin.") in self.run_keyword_variants:
            col = keyword_col(node)
            self.report("if-can-be-used", run_keyword=node.keyword, node=node, col=col)


class ConsistentAssignmentSignChecker(VisitorChecker):
    """Checker for inconsistent assignment signs.

    By default this checker will try to autodetect most common assignment sign (separately for *** Variables *** section
    and (*** Test Cases ***, *** Keywords ***) sections and report any not consistent type of sign in particular file.

    To force one type of sign type you, can configure two rules::

        --configure inconsistent-assignment:assignment_sign_type:{sign_type}
        --configure inconsistent-assignment-in-variables:assignment_sign_type:{sign_type}

    ``${sign_type}`` can be one of: ``autodetect`` (default), ``none`` (''), ``equal_sign`` ('='),
    ``space_and_equal_sign`` (' =').

    """

    reports = (
        "inconsistent-assignment",
        "inconsistent-assignment-in-variables",
    )

    def __init__(self):
        self.keyword_expected_sign_type = None
        self.variables_expected_sign_type = None
        super().__init__()

    def visit_File(self, node):  # noqa
        self.keyword_expected_sign_type = self.param("inconsistent-assignment", "assignment_sign_type")
        self.variables_expected_sign_type = self.param("inconsistent-assignment-in-variables", "assignment_sign_type")
        if "autodetect" in [
            self.param("inconsistent-assignment", "assignment_sign_type"),
            self.param("inconsistent-assignment-in-variables", "assignment_sign_type"),
        ]:
            auto_detector = self.auto_detect_assignment_sign(node)
            if self.param("inconsistent-assignment", "assignment_sign_type") == "autodetect":
                self.keyword_expected_sign_type = auto_detector.keyword_most_common
            if self.param("inconsistent-assignment-in-variables", "assignment_sign_type") == "autodetect":
                self.variables_expected_sign_type = auto_detector.variables_most_common
        self.generic_visit(node)

    def visit_KeywordCall(self, node):  # noqa
        if self.keyword_expected_sign_type is None or not node.keyword:
            return
        if node.assign:  # if keyword returns any value
            assign_tokens = node.get_tokens(Token.ASSIGN)
            self.check_assign_type(
                assign_tokens[-1],
                self.keyword_expected_sign_type,
                "inconsistent-assignment",
            )
        return node

    def visit_VariableSection(self, node):  # noqa
        if self.variables_expected_sign_type is None:
            return
        for child in node.body:
            if not isinstance(child, Variable) or get_errors(child):
                continue
            var_token = child.get_token(Token.VARIABLE)
            self.check_assign_type(
                var_token,
                self.variables_expected_sign_type,
                "inconsistent-assignment-in-variables",
            )
        return node

    def check_assign_type(self, token, expected, issue_name):
        sign = AssignmentTypeDetector.get_assignment_sign(token.value)
        if sign != expected:
            self.report(
                issue_name,
                expected_sign=expected,
                actual_sign=sign,
                lineno=token.lineno,
                col=token.end_col_offset + 1,
            )

    @staticmethod
    def auto_detect_assignment_sign(node):
        auto_detector = AssignmentTypeDetector()
        auto_detector.visit(node)
        return auto_detector


class SettingsOrderChecker(VisitorChecker):
    """Checker for settings order.

    BuiltIn libraries imports should always be placed before other libraries imports.
    """

    reports = ("wrong-import-order",)

    def __init__(self):
        self.libraries = []
        super().__init__()

    def visit_File(self, node):  # noqa
        self.libraries = []
        self.generic_visit(node)
        first_non_builtin = None
        for library in self.libraries:
            if first_non_builtin is None:
                if library.name not in STDLIBS:
                    first_non_builtin = library.name
            else:
                if library.name in STDLIBS:
                    self.report(
                        "wrong-import-order",
                        builtin_import=library.name,
                        custom_import=first_non_builtin,
                        node=library,
                    )

    def visit_LibraryImport(self, node):  # noqa
        if not node.name:
            return
        self.libraries.append(node)


class EmptyVariableChecker(VisitorChecker):
    """Checker for variables without value."""

    reports = ("empty-variable",)

    def visit_Variable(self, node):  # noqa
        if get_errors(node):
            return
        if not node.value:  # catch variable declaration without any value
            self.report("empty-variable", node=node)
        for token in node.get_tokens(Token.ARGUMENT):
            if not token.value or token.value == "\\":
                self.report(
                    "empty-variable",
                    node=token,
                    lineno=token.lineno,
                    col=token.col_offset,
                )


class ResourceFileChecker(VisitorChecker):
    """Checker for resource files."""

    reports = ("can-be-resource-file",)

    def visit_File(self, node):  # noqa
        source = node.source if node.source else self.source
        if source:
            extension = Path(source).suffix
            file_name = Path(source).stem
            if (
                ".resource" not in extension
                and "__init__" not in file_name
                and node.sections
                and not any([isinstance(section, TestCaseSection) for section in node.sections])
            ):
                self.report("can-be-resource-file", file_name=Path(source).name, file_name_stem=file_name, node=node)


class IfChecker(VisitorChecker):
    """Checker for IF blocks"""

    reports = (
        "if-can-be-merged",
        "inline-if-can-be-used",
    )

    def visit_TestCase(self, node):  # noqa
        if get_errors(node):
            return
        self.check_adjacent_ifs(node)

    visit_For = visit_If = visit_Keyword = visit_TestCase  # TODO  While, Try Except?

    @staticmethod
    def is_if_inline(node):
        return isinstance(node.header, InlineIfHeader)

    def check_adjacent_ifs(self, node):
        previous_if = None
        for child in node.body:
            if isinstance(child, If):
                if child.header.errors:
                    continue
                self.check_whether_if_should_be_inline(child)
                if previous_if and child.header and self.compare_conditions(child, previous_if):
                    token = child.header.get_token(child.header.type)
                    self.report("if-can-be-merged", line=previous_if.lineno, node=token, col=token.col_offset + 1)
                previous_if = child
            elif not isinstance(child, (Comment, EmptyLine)):
                previous_if = None
        self.generic_visit(node)

    def compare_conditions(self, if_node, other_if_node):
        if not self.compare_assign_tokens(if_node, other_if_node):
            return False
        while if_node is not None and other_if_node is not None:
            if if_node.condition != other_if_node.condition:
                return False
            if_node = if_node.orelse
            other_if_node = other_if_node.orelse
        return if_node is None and other_if_node is None

    @staticmethod
    def normalize_var_name(name):
        return name.lower().replace("_", "").replace(" ", "").replace("=", "")

    def compare_assign_tokens(self, if_node, other_if_node):
        assign_1 = getattr(if_node, "assign", None)
        assign_2 = getattr(other_if_node, "assign", None)
        if assign_1 is None or assign_2 is None:
            return all(assign is None for assign in (assign_1, assign_2))
        if len(assign_1) != len(assign_2):
            return False
        for var1, var2 in zip(assign_1, assign_2):
            if self.normalize_var_name(var1) != self.normalize_var_name(var2):
                return False
        return True

    @staticmethod
    def tokens_length(tokens):
        return sum(len(token.value) for token in tokens)

    def check_whether_if_should_be_inline(self, node):
        if ROBOT_VERSION.major < 5:
            return
        if self.is_if_inline(node):
            return
        if (
            len(node.body) != 1
            or node.orelse
            or not isinstance(node.body[0], (KeywordCall, ReturnStatement, Break, Continue))  # type: ignore[arg-type]
        ):
            return
        min_possible = self.tokens_length(node.header.tokens) + self.tokens_length(node.body[0].tokens[1:]) + 2
        if min_possible > self.param("inline-if-can-be-used", "max_width"):
            return
        token = node.header.get_token(node.header.type)
        self.report("inline-if-can-be-used", node=node, col=token.col_offset + 1)


class LoopStatementsChecker(VisitorChecker):
    """Checker for loop keywords and statements such as CONTINUE or Exit For Loop"""

    reports = ("statement-outside-loop",)
    for_keyword = {"continueforloop", "continueforloopif", "exitforloop", "exitforloopif"}

    def __init__(self):
        self.loops = 0
        super().__init__()

    def visit_File(self, node):  # noqa
        self.loops = 0
        self.generic_visit(node)

    def visit_For(self, node):  # noqa
        self.loops += 1
        self.generic_visit(node)
        self.loops -= 1

    visit_While = visit_For

    def visit_KeywordCall(self, node):  # noqa
        if node.errors or self.loops:
            return
        if normalize_robot_name(node.keyword, remove_prefix="builtin.") in self.for_keyword:
            self.report(
                "statement-outside-loop",
                name=f"'{node.keyword}'",
                statement_type="keyword",
                node=node,
                col=keyword_col(node),
            )

    def visit_Continue(self, node):  # noqa
        self.check_statement_in_loop(node, "CONTINUE")  # type: ignore[arg-type]

    def visit_Break(self, node):  # noqa
        self.check_statement_in_loop(node, "BREAK")  # type: ignore[arg-type]

    def check_statement_in_loop(self, node, token_type):
        if self.loops or node.errors and f"{token_type} can only be used inside a loop." not in node.errors:
            return
        self.report(
            "statement-outside-loop",
            name=token_type,
            statement_type="statement",
            node=node,
            col=token_col(node, token_type),
        )



# fmt: off
# Warning: Don't edit file (autogenerated from python -m dev codegen).
from typing import List

ROBOT_RUN_TEST = "robot.runTest"  # Run Test/Task
ROBOT_DEBUG_TEST = "robot.debugTest"  # Debug Test/Task
ROBOT_RUN_SUITE = "robot.runSuite"  # Run Tests/Tasks Suite
ROBOT_DEBUG_SUITE = "robot.debugSuite"  # Debug Tests/Tasks Suite
ROBOT_INTERACTIVE_SHELL = "robot.interactiveShell"  # Start Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_START = "robot.internal.rfinteractive.start"  # Create Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE = "robot.internal.rfinteractive.evaluate"  # Evaluate in Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_STOP = "robot.internal.rfinteractive.stop"  # Stop Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS = "robot.internal.rfinteractive.semanticTokens"  # Get the semantic tokens based on the code entered.
ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION = "robot.internal.rfinteractive.resolveCompletion"  # Resolves the passed completion.
ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS = "robot.internal.rfinteractive.completions"  # Get the completions based on the code entered.
ROBOT_GET_RFLS_HOME_DIR = "robot.getRFLSHomeDir"  # Provides the directory used to store information (usually ~/.robotframework-ls -- may be configured with `ROBOTFRAMEWORK_LS_USER_HOME` environment variable).
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES = "robot.clearCachesAndRestartProcesses"  # Clear caches and restart Robot Framework Language Server
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_START_INTERNAL = "robot.clearCachesAndRestartProcesses.start.internal"  # Stops the RFLS and waits for robot.clearCachesAndRestartProcesses.finish.internal to restart
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_FINISH_INTERNAL = "robot.clearCachesAndRestartProcesses.finish.internal"  # To be used to restart the processes
ROBOT_START_INDEXING_INTERNAL = "robot.startIndexing.internal"  # Starts the indexing service
ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL = "robot.waitFullTestCollection.internal"  # Schedules and Waits for a full test collection

ALL_SERVER_COMMANDS: List[str] = [
    ROBOT_INTERNAL_RFINTERACTIVE_START,
    ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE,
    ROBOT_INTERNAL_RFINTERACTIVE_STOP,
    ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS,
    ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION,
    ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS,
    ROBOT_GET_RFLS_HOME_DIR,
    ROBOT_START_INDEXING_INTERNAL,
    ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL,
]

# fmt: on



import os
import enum
from typing import Optional

LOG_FILENAME: Optional[str] = os.getenv("ROBOTFRAMEWORK_DAP_LOG_FILENAME", None)

# Make sure that the log level is an int.
try:
    LOG_LEVEL = int(os.getenv("ROBOTFRAMEWORK_DAP_LOG_LEVEL", "0"))
except:
    LOG_LEVEL = 3

DEBUG = LOG_LEVEL > 1

TERMINAL_NONE = "none"
TERMINAL_INTEGRATED = "integrated"
TERMINAL_EXTERNAL = "external"

VALID_TERMINAL_OPTIONS = [TERMINAL_NONE, TERMINAL_INTEGRATED, TERMINAL_EXTERNAL]


STATE_RUNNING = "running"
STATE_PAUSED = "paused"

# See: StoppedEvent


class ReasonEnum(enum.Enum):
    REASON_NOT_STOPPED = "not_stopped"
    REASON_BREAKPOINT = "breakpoint"
    REASON_STEP = "step"
    REASON_PAUSE = "pause"
    REASON_EXCEPTION = "exception"


class StepEnum(enum.Enum):
    STEP_NONE = 0
    STEP_IN = 1
    STEP_NEXT = 2
    STEP_OUT = 3



from zipped import zipped_contents
try:
    code = zipped_contents.call_in_zip.__code__
except AttributeError:
    code = zipped_contents.call_in_zip.func_code
assert 'myzip.zip' in code.co_filename
zipped_contents.call_in_zip()

from zipped2 import zipped_contents2
try:
    code = zipped_contents2.call_in_zip2.__code__
except AttributeError:
    code = zipped_contents2.call_in_zip2.func_code
assert 'myzip2.egg!' in code.co_filename
zipped_contents2.call_in_zip2()

print('TEST SUCEEDED!')



"""
Lengths checkers
"""
import re

from robot.parsing.model.blocks import CommentSection, TestCase
from robot.parsing.model.statements import Arguments, Comment, EmptyLine, KeywordCall

from robocop.checkers import RawFileChecker, VisitorChecker
from robocop.rules import Rule, RuleParam, RuleSeverity
from robocop.utils import last_non_empty_line, normalize_robot_name, pattern_type, get_section_name

rules = {
    "0501": Rule(
        RuleParam(name="max_len", default=40, converter=int, desc="number of lines allowed in a keyword"),
        rule_id="0501",
        name="too-long-keyword",
        msg="Keyword '{{ keyword_name }}' is too long ({{ keyword_length }}/{{ allowed_length}})",
        severity=RuleSeverity.WARNING,
    ),
    "0502": Rule(
        RuleParam(name="min_calls", default=1, converter=int, desc="number of keyword calls required in a keyword"),
        rule_id="0502",
        name="too-few-calls-in-keyword",
        msg="Keyword '{{ keyword_name }}' has too few keywords inside ({{ keyword_count }}/{{ min_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0503": Rule(
        RuleParam(name="max_calls", default=10, converter=int, desc="number of keyword calls allowed in a keyword"),
        rule_id="0503",
        name="too-many-calls-in-keyword",
        msg="Keyword '{{ keyword_name }}' has too many keywords inside ({{ keyword_count }}/{{ max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0504": Rule(
        RuleParam(name="max_len", default=20, converter=int, desc="number of lines allowed in a test case"),
        rule_id="0504",
        name="too-long-test-case",
        msg="Test case '{{ test_name }}' is too long ({{ test_length }}/{{ allowed_length }})",
        severity=RuleSeverity.WARNING,
    ),
    "0505": Rule(
        RuleParam(name="max_calls", default=10, converter=int, desc="number of keyword calls allowed in a test case"),
        rule_id="0505",
        name="too-many-calls-in-test-case",
        msg="Test case '{{ test_name }}' has too many keywords inside ({{ keyword_count }}/{{ max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0506": Rule(
        RuleParam(name="max_lines", default=400, converter=int, desc="number of lines allowed in a file"),
        rule_id="0506",
        name="file-too-long",
        msg="File has too many lines ({{ lines_count }}/{{max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0507": Rule(
        RuleParam(name="max_args", default=5, converter=int, desc="number of lines allowed in a file"),
        rule_id="0507",
        name="too-many-arguments",
        msg="Keyword '{{ keyword_name }}' has too many arguments ({{ arguments_count }}/{{ max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0508": Rule(
        RuleParam(name="line_length", default=120, converter=int, desc="number of characters allowed in line"),
        RuleParam(
            name="ignore_pattern",
            default=re.compile(r"https?://\S+"),
            converter=pattern_type,
            desc="ignore lines that contain configured pattern",
        ),
        rule_id="0508",
        name="line-too-long",
        msg="Line is too long ({{ line_length }}/{{ allowed_length }})",
        severity=RuleSeverity.WARNING,
        docs="""
        It is possible to ignore lines that match regex pattern. Configure it using following option::
        
            robocop --configure line-too-long:ignore_pattern:pattern
        
        The default pattern is `https?://\S+` that ignores the lines that look like an URL.

        """,
    ),
    "0509": Rule(
        rule_id="0509", name="empty-section", msg="Section '{{ section_name }}' is empty", severity=RuleSeverity.WARNING
    ),
    "0510": Rule(
        RuleParam(
            name="max_returns", default=4, converter=int, desc="allowed number of returned values from a keyword"
        ),
        rule_id="0510",
        name="number-of-returned-values",
        msg="Too many return values ({{ return_count }}/{{ max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
    "0511": Rule(
        rule_id="0511",
        name="empty-metadata",
        msg="Metadata settings does not have any value set",
        severity=RuleSeverity.WARNING,
    ),
    "0512": Rule(
        rule_id="0512",
        name="empty-documentation",
        msg="Documentation of {{ block_name }} is empty",
        severity=RuleSeverity.WARNING,
    ),
    "0513": Rule(rule_id="0513", name="empty-force-tags", msg="Force Tags are empty", severity=RuleSeverity.WARNING),
    "0514": Rule(
        rule_id="0514", name="empty-default-tags", msg="Default Tags are empty", severity=RuleSeverity.WARNING
    ),
    "0515": Rule(
        rule_id="0515", name="empty-variables-import", msg="Import variables path is empty", severity=RuleSeverity.ERROR
    ),
    "0516": Rule(
        rule_id="0516", name="empty-resource-import", msg="Import resource path is empty", severity=RuleSeverity.ERROR
    ),
    "0517": Rule(
        rule_id="0517", name="empty-library-import", msg="Import library path is empty", severity=RuleSeverity.ERROR
    ),
    "0518": Rule(
        rule_id="0518",
        name="empty-setup",
        msg="Setup of {{ block_name }} does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0519": Rule(
        rule_id="0519",
        name="empty-suite-setup",
        msg="Suite Setup does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0520": Rule(
        rule_id="0520",
        name="empty-test-setup",
        msg="Test Setup does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0521": Rule(
        rule_id="0521",
        name="empty-teardown",
        msg="Teardown of {{ block_name }} does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0522": Rule(
        rule_id="0522",
        name="empty-suite-teardown",
        msg="Suite Teardown does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0523": Rule(
        rule_id="0523",
        name="empty-test-teardown",
        msg="Test Teardown does not have any keywords",
        severity=RuleSeverity.ERROR,
    ),
    "0524": Rule(
        rule_id="0524", name="empty-timeout", msg="Timeout of {{ block_name }} is empty", severity=RuleSeverity.WARNING
    ),
    "0525": Rule(rule_id="0525", name="empty-test-timeout", msg="Test Timeout is empty", severity=RuleSeverity.WARNING),
    "0526": Rule(
        rule_id="0526",
        name="empty-arguments",
        msg="Arguments of {{ block_name }} are empty",
        severity=RuleSeverity.ERROR,
    ),
    "0527": Rule(
        RuleParam(name="max_testcases", default=50, converter=int, desc="number of test cases allowed in a suite"),
        RuleParam(
            name="max_templated_testcases",
            default=100,
            converter=int,
            desc="number of test cases allowed in a templated suite",
        ),
        rule_id="0527",
        name="too-many-test-cases",
        msg="Too many test cases ({{ test_count }}/{{ max_allowed_count }})",
        severity=RuleSeverity.WARNING,
    ),
}


class LengthChecker(VisitorChecker):
    """Checker for max and min length of keyword or test case. It analyses number of lines and also number of
    keyword calls (as you can have just few keywords but very long ones or vice versa).
    """

    reports = (
        "too-long-keyword",
        "too-few-calls-in-keyword",
        "too-many-calls-in-keyword",
        "too-long-test-case",
        "too-many-calls-in-test-case",
        "file-too-long",
        "too-many-arguments",
    )

    def visit_File(self, node):
        if node.end_lineno > self.param("file-too-long", "max_lines"):
            self.report(
                "file-too-long",
                lines_count=node.end_lineno,
                max_allowed_count=self.param("file-too-long", "max_lines"),
                node=node,
                lineno=node.end_lineno,
            )
        super().visit_File(node)

    def visit_Keyword(self, node):  # noqa
        if node.name.lstrip().startswith("#"):
            return
        for child in node.body:
            if isinstance(child, Arguments):
                args_number = len(child.values)
                if args_number > self.param("too-many-arguments", "max_args"):
                    self.report(
                        "too-many-arguments",
                        keyword_name=node.name,
                        arguments_count=args_number,
                        max_allowed_count=self.param("too-many-arguments", "max_args"),
                        node=node,
                    )
                break
        length = LengthChecker.check_node_length(node)
        if length > self.param("too-long-keyword", "max_len"):
            self.report(
                "too-long-keyword",
                keyword_name=node.name,
                keyword_length=length,
                allowed_length=self.param("too-long-keyword", "max_len"),
                node=node,
                lineno=node.end_lineno,
                ext_disablers=(node.lineno, last_non_empty_line(node)),
            )
            return
        key_calls = LengthChecker.count_keyword_calls(node)
        if key_calls < self.param("too-few-calls-in-keyword", "min_calls"):
            self.report(
                "too-few-calls-in-keyword",
                keyword_name=node.name,
                keyword_count=key_calls,
                min_allowed_count=self.param("too-few-calls-in-keyword", "min_calls"),
                node=node,
            )
            return
        if key_calls > self.param("too-many-calls-in-keyword", "max_calls"):
            self.report(
                "too-many-calls-in-keyword",
                keyword_name=node.name,
                keyword_count=key_calls,
                max_allowed_count=self.param("too-many-calls-in-keyword", "max_calls"),
                node=node,
            )
            return

    def visit_TestCase(self, node):  # noqa
        length = LengthChecker.check_node_length(node)
        if length > self.param("too-long-test-case", "max_len"):
            self.report(
                "too-long-test-case",
                test_name=node.name,
                test_length=length,
                allowed_length=self.param("too-long-test-case", "max_len"),
                node=node,
            )
        key_calls = LengthChecker.count_keyword_calls(node)
        if key_calls > self.param("too-many-calls-in-test-case", "max_calls"):
            self.report(
                "too-many-calls-in-test-case",
                test_name=node.name,
                keyword_count=key_calls,
                max_allowed_count=self.param("too-many-calls-in-test-case", "max_calls"),
                node=node,
            )
            return

    @staticmethod
    def check_node_length(node):
        return node.end_lineno - node.lineno

    @staticmethod
    def count_keyword_calls(node):
        if isinstance(node, KeywordCall):
            return 1
        if not hasattr(node, "body"):
            return 0
        return sum(LengthChecker.count_keyword_calls(child) for child in node.body)


class LineLengthChecker(RawFileChecker):
    """Checker for maximum length of a line."""

    reports = ("line-too-long",)
    # replace `# noqa` or `# robocop`, `# robocop: enable`, `# robocop: disable=optional,rule,names`
    disabler_pattern = re.compile(r"(# )+(noqa|robocop: ?(?P<disabler>disable|enable)=?(?P<rules>[\w\-,]*))")

    def check_line(self, line, lineno):
        if self.param("line-too-long", "ignore_pattern") and self.param("line-too-long", "ignore_pattern").search(line):
            return
        line = self.disabler_pattern.sub("", line)
        line = line.rstrip().expandtabs(4)
        if len(line) > self.param("line-too-long", "line_length"):
            self.report(
                "line-too-long",
                line_length=len(line),
                allowed_length=self.param("line-too-long", "line_length"),
                lineno=lineno,
            )


class EmptySectionChecker(VisitorChecker):
    """Checker for detecting empty sections."""

    reports = ("empty-section",)

    def check_if_empty(self, node):
        if not node.header:
            return
        anything_but = EmptyLine if isinstance(node, CommentSection) else (Comment, EmptyLine)
        if all(isinstance(child, anything_but) for child in node.body):
            self.report("empty-section", section_name=get_section_name(node), node=node)

    def visit_Section(self, node):  # noqa
        self.check_if_empty(node)


class NumberOfReturnedArgsChecker(VisitorChecker):
    """Checker for number of returned values from a keyword."""

    reports = ("number-of-returned-values",)

    def visit_Return(self, node):  # noqa
        self.check_node_returns(len(node.values), node)

    visit_ReturnStatement = visit_Return

    def visit_KeywordCall(self, node):  # noqa
        if not node.keyword:
            return

        normalized_name = normalize_robot_name(node.keyword, remove_prefix="builtin.")
        if normalized_name == "returnfromkeyword":
            self.check_node_returns(len(node.args), node)
        elif normalized_name == "returnfromkeywordif":
            self.check_node_returns(len(node.args) - 1, node)

    def check_node_returns(self, return_count, node):
        if return_count > self.param("number-of-returned-values", "max_returns"):
            self.report(
                "number-of-returned-values",
                return_count=return_count,
                max_allowed_count=self.param("number-of-returned-values", "max_returns"),
                node=node,
                col=node.data_tokens[0].col_offset + 1,
            )


class EmptySettingsChecker(VisitorChecker):
    """Checker for detecting empty settings."""

    reports = (
        "empty-metadata",
        "empty-documentation",
        "empty-force-tags",
        "empty-default-tags",
        "empty-variables-import",
        "empty-resource-import",
        "empty-library-import",
        "empty-setup",
        "empty-suite-setup",
        "empty-test-setup",
        "empty-teardown",
        "empty-suite-teardown",
        "empty-test-teardown",
        "empty-timeout",
        "empty-test-timeout",
        "empty-arguments",
    )

    def __init__(self):
        self.parent_node_name = ""
        super().__init__()

    def visit_SettingSection(self, node):  # noqa
        self.parent_node_name = "Test Suite"
        self.generic_visit(node)

    def visit_TestCaseName(self, node):  # noqa
        if node.name:
            self.parent_node_name = f"'{node.name}' Test Case"
        else:
            self.parent_node_name = ""
        self.generic_visit(node)

    def visit_Keyword(self, node):  # noqa
        if node.name:
            self.parent_node_name = f"'{node.name}' Keyword"
        else:
            self.parent_node_name = ""
        self.generic_visit(node)

    def visit_Metadata(self, node):  # noqa
        if node.name is None:
            self.report("empty-metadata", node=node, col=node.end_col_offset)

    def visit_Documentation(self, node):  # noqa
        if not node.value:
            self.report("empty-documentation", block_name=self.parent_node_name, node=node, col=node.end_col_offset)

    def visit_ForceTags(self, node):  # noqa
        if not node.values:
            self.report("empty-force-tags", node=node, col=node.end_col_offset)

    def visit_DefaultTags(self, node):  # noqa
        if not node.values:
            self.report("empty-default-tags", node=node, col=node.end_col_offset)

    def visit_VariablesImport(self, node):  # noqa
        if not node.name:
            self.report("empty-variables-import", node=node, col=node.end_col_offset)

    def visit_ResourceImport(self, node):  # noqa
        if not node.name:
            self.report("empty-resource-import", node=node, col=node.end_col_offset)

    def visit_LibraryImport(self, node):  # noqa
        if not node.name:
            self.report("empty-library-import", node=node, col=node.end_col_offset)

    def visit_Setup(self, node):  # noqa
        if not node.name:
            self.report("empty-setup", block_name=self.parent_node_name, node=node, col=node.end_col_offset + 1)

    def visit_SuiteSetup(self, node):  # noqa
        if not node.name:
            self.report("empty-suite-setup", node=node, col=node.end_col_offset)

    def visit_TestSetup(self, node):  # noqa
        if not node.name:
            self.report("empty-test-setup", node=node, col=node.end_col_offset)

    def visit_Teardown(self, node):  # noqa
        if not node.name:
            self.report("empty-teardown", block_name=self.parent_node_name, node=node, col=node.end_col_offset + 1)

    def visit_SuiteTeardown(self, node):  # noqa
        if not node.name:
            self.report("empty-suite-teardown", node=node, col=node.end_col_offset)

    def visit_TestTeardown(self, node):  # noqa
        if not node.name:
            self.report("empty-test-teardown", node=node, col=node.end_col_offset)

    def visit_Timeout(self, node):  # noqa
        if not node.value:
            self.report("empty-timeout", block_name=self.parent_node_name, node=node, col=node.end_col_offset + 1)

    def visit_TestTimeout(self, node):  # noqa
        if not node.value:
            self.report("empty-test-timeout", node=node, col=node.end_col_offset)

    def visit_Arguments(self, node):  # noqa
        if not node.values:
            self.report("empty-arguments", block_name=self.parent_node_name, node=node, col=node.end_col_offset + 1)


class TestCaseNumberChecker(VisitorChecker):
    """Checker for counting number of test cases depending on suite type"""

    reports = ("too-many-test-cases",)

    def visit_TestCaseSection(self, node):  # noqa
        max_testcases = (
            self.param("too-many-test-cases", "max_templated_testcases")
            if self.templated_suite
            else self.param("too-many-test-cases", "max_testcases")
        )
        discovered_testcases = sum([isinstance(child, TestCase) for child in node.body])
        if discovered_testcases > max_testcases:
            self.report(
                "too-many-test-cases", test_count=discovered_testcases, max_allowed_count=max_testcases, node=node
            )



import sys
import functools

import click

from robotidy.utils import node_within_lines
from robotidy.exceptions import (
    InvalidParameterValueError,
    InvalidParameterError,
    InvalidParameterFormatError,
    ImportTransformerError,
)


def check_start_end_line(func):
    """
    Do not transform node if it's not within passed start_line and end_line.
    """

    @functools.wraps(func)
    def wrapper(self, node, *args):
        if not node:
            return node
        if not node_within_lines(
            node.lineno,
            node.end_lineno,
            self.formatting_config.start_line,
            self.formatting_config.end_line,
        ):
            return node
        return func(self, node, *args)

    return wrapper


def catch_exceptions(func):
    """
    Catch exceptions and print user friendly message for common issues
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not func:
            return functools.partial(catch_exceptions)
        try:
            return func(*args, **kwargs)
        except (
            InvalidParameterValueError,
            InvalidParameterError,
            InvalidParameterFormatError,
            ImportTransformerError,
        ) as err:
            print(f"Error: {err}")
            sys.exit(1)
        except (click.exceptions.ClickException, click.exceptions.Exit):
            raise
        except Exception as err:
            message = (
                "\nFatal exception occurred. You can create an issue at "
                "https://github.com/MarketSquare/robotframework-tidy/issues . Thanks!"
            )
            err.args = (err.args[0] + message,) + err.args[1:]
            raise err

    return wrapper
